import torch
from diffusers import HunyuanVideoPipeline, HunyuanVideoTransformer3DModel
from diffusers.utils import export_to_video

model_id = "tencent/HunyuanVideo"
transformer = HunyuanVideoTransformer3DModel.from_pretrained(
    model_id,
    subfolder="transformer",
    torch_dtype=torch.bfloat16,
    revision="refs/pr/18",
)
pipe = HunyuanVideoPipeline.from_pretrained(
    model_id,
    transformer=transformer,
    torch_dtype=torch.float16,
    revision="refs/pr/18",
).to("cuda")

from para_attn.first_block_cache.diffusers_adapters import apply_cache_on_pipe

apply_cache_on_pipe(pipe)

# Enable memory savings
# pipe.enable_model_cpu_offload()
pipe.vae.enable_tiling(
    # Make it runnable on GPUs with 48GB memory
    # tile_sample_min_height=128,
    # tile_sample_stride_height=96,
    # tile_sample_min_width=128,
    # tile_sample_stride_width=96,
    # tile_sample_min_num_frames=32,
    # tile_sample_stride_num_frames=24,
)

# pipe.transformer = torch.compile(pipe.transformer, mode="max-autotune-no-cudagraphs")

output = pipe(
    prompt="A cat walks on the grass, realistic",
    height=720,
    width=1280,
    num_frames=129,
    num_inference_steps=30,
).frames[0]

print("Saving video to hunyuan_video.mp4")
export_to_video(output, "hunyuan_video.mp4", fps=15)



import torch
from diffusers import MochiPipeline
from diffusers.utils import export_to_video

pipe = MochiPipeline.from_pretrained(
    "genmo/mochi-1-preview",
    torch_dtype=torch.bfloat16,
).to("cuda")

from para_attn.first_block_cache.diffusers_adapters import apply_cache_on_pipe

apply_cache_on_pipe(pipe)

# Enable memory savings
# pipe.enable_model_cpu_offload()
pipe.enable_vae_tiling()

# pipe.transformer = torch.compile(pipe.transformer, mode="max-autotune-no-cudagraphs")

prompt = "Close-up of a chameleon's eye, with its scaly skin changing color. Ultra high resolution 4k."
video = pipe(
    prompt,
    num_frames=84,
).frames[0]

print("Saving video to mochi.mp4")
export_to_video(video, "mochi.mp4", fps=30)



import functools
from typing import List, Optional, Union

import torch
from diffusers import DiffusionPipeline, FluxTransformer2DModel

import para_attn.primitives as DP
from para_attn.context_parallel import init_context_parallel_mesh
from para_attn.para_attn_interface import UnifiedAttnMode


def parallelize_transformer(transformer: FluxTransformer2DModel, *, mesh=None):
    if getattr(transformer, "_is_parallelized", False):
        return transformer

    mesh = init_context_parallel_mesh(transformer.device.type, mesh=mesh)
    batch_mesh = mesh["batch"]
    seq_mesh = mesh["ring", "ulysses"]._flatten()

    original_forward = transformer.forward

    @functools.wraps(transformer.__class__.forward)
    def new_forward(
        self,
        hidden_states: torch.Tensor,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        *args,
        timestep: torch.LongTensor = None,
        img_ids: torch.Tensor = None,
        txt_ids: torch.Tensor = None,
        controlnet_block_samples: Optional[List[torch.Tensor]] = None,
        controlnet_single_block_samples: Optional[List[torch.Tensor]] = None,
        **kwargs,
    ):
        if isinstance(timestep, torch.Tensor) and timestep.ndim != 0 and timestep.shape[0] == hidden_states.shape[0]:
            timestep = DP.get_assigned_chunk(timestep, dim=0, group=batch_mesh)
        hidden_states = DP.get_assigned_chunk(hidden_states, dim=0, group=batch_mesh)
        hidden_states = DP.get_assigned_chunk(hidden_states, dim=-2, group=seq_mesh)
        encoder_hidden_states = DP.get_assigned_chunk(encoder_hidden_states, dim=0, group=batch_mesh)
        encoder_hidden_states = DP.get_assigned_chunk(encoder_hidden_states, dim=-2, group=seq_mesh)
        img_ids = DP.get_assigned_chunk(img_ids, dim=-2)
        txt_ids = DP.get_assigned_chunk(txt_ids, dim=-2)
        if controlnet_block_samples is not None:
            controlnet_block_samples = [
                DP.get_assigned_chunk(sample, dim=0, group=batch_mesh) for sample in controlnet_block_samples
            ]
            controlnet_block_samples = [
                DP.get_assigned_chunk(sample, dim=-2, group=seq_mesh) for sample in controlnet_block_samples
            ]
            kwargs["controlnet_block_samples"] = controlnet_block_samples
        if controlnet_single_block_samples is not None:
            controlnet_single_block_samples = [
                DP.get_assigned_chunk(sample, dim=0, group=batch_mesh) for sample in controlnet_single_block_samples
            ]
            controlnet_single_block_samples = [
                DP.get_assigned_chunk(sample, dim=-2, group=seq_mesh) for sample in controlnet_single_block_samples
            ]
            kwargs["controlnet_single_block_samples"] = controlnet_single_block_samples

        with UnifiedAttnMode(mesh):
            output = original_forward(
                hidden_states,
                encoder_hidden_states,
                *args,
                timestep=timestep,
                img_ids=img_ids,
                txt_ids=txt_ids,
                **kwargs,
            )

        return_dict = not isinstance(output, tuple)
        sample = output[0]
        sample = DP.get_complete_tensor(sample, dim=-2, group=seq_mesh)
        sample = DP.get_complete_tensor(sample, dim=0, group=batch_mesh)
        if return_dict:
            return output.__class__(sample, *output[1:])
        return (sample, *output[1:])

    transformer.forward = new_forward.__get__(transformer)

    transformer._is_parallelized = True

    return transformer


def parallelize_pipe(pipe: DiffusionPipeline, *, shallow_patch: bool = False, **kwargs):
    if not getattr(pipe, "_is_parallelized", False):
        original_call = pipe.__class__.__call__

        @functools.wraps(original_call)
        def new_call(self, *args, generator: Optional[Union[torch.Generator, List[torch.Generator]]] = None, **kwargs):
            if generator is None and getattr(self, "_is_parallelized", False):
                seed_t = torch.randint(0, torch.iinfo(torch.int64).max, [1], dtype=torch.int64, device=self.device)
                seed_t = DP.get_complete_tensor(seed_t, dim=0)
                seed_t = DP.get_assigned_chunk(seed_t, dim=0, idx=0)
                seed = seed_t.item()
                seed -= torch.iinfo(torch.int64).min
                generator = torch.Generator(self.device).manual_seed(seed)
            return original_call(self, *args, generator=generator, **kwargs)

        new_call._is_parallelized = True

        pipe.__class__.__call__ = new_call
        pipe.__class__._is_parallelized = True

    if not shallow_patch:
        parallelize_transformer(pipe.transformer, **kwargs)

    return pipe



from typing import List, Optional, Tuple, Union

import torch
import torch.distributed as dist

if dist.is_available():
    import torch.distributed._functional_collectives as ft_c
    import torch.distributed.distributed_c10d as c10d
else:
    ft_c = None
    c10d = None


def get_group(group=None):
    if group is None:
        group = c10d._get_default_group()

    if isinstance(group, dist.ProcessGroup):
        pg: Union[dist.ProcessGroup, List[dist.ProcessGroup]] = group
    else:
        pg = group.get_group()

    return pg


def get_world_size(group=None):
    pg = get_group(group)
    return dist.get_world_size(pg)


def get_rank(group=None):
    pg = get_group(group)
    return dist.get_rank(pg)


def _maybe_wait(tensor: torch.Tensor) -> torch.Tensor:
    """
    When tracing the code, the result tensor is not an AsyncCollectiveTensor,
    so we cannot call ``wait()``.
    """
    if isinstance(tensor, ft_c.AsyncCollectiveTensor):
        return tensor.wait()
    return tensor


def all_gather_tensor_sync(x, *args, group=None, **kwargs):
    group = get_group(group)
    x_shape = x.shape
    x = x.flatten()
    x_numel = x.numel()
    x = ft_c.all_gather_tensor(x, *args, group=group, **kwargs)
    x = _maybe_wait(x)
    x_shape = list(x_shape)
    x_shape[0] *= x.numel() // x_numel
    x = x.reshape(x_shape)
    return x


def all_gather_tensor_autograd_sync(x, *args, group=None, **kwargs):
    group = get_group(group)
    x_shape = x.shape
    x = x.flatten()
    x_numel = x.numel()
    x = ft_c.all_gather_tensor_autograd(x, *args, group=group, **kwargs)
    x = _maybe_wait(x)
    x_shape = list(x_shape)
    x_shape[0] *= x.numel() // x_numel
    x = x.reshape(x_shape)
    return x


def all_to_all_single_sync(x, *args, **kwargs):
    x_shape = x.shape
    x = x.flatten()
    x = ft_c.all_to_all_single(x, *args, **kwargs)
    x = _maybe_wait(x)
    x = x.reshape(x_shape)
    return x


def all_to_all_single_autograd_sync(x, *args, **kwargs):
    x_shape = x.shape
    x = x.flatten()
    x = ft_c.all_to_all_single_autograd(x, *args, **kwargs)
    x = _maybe_wait(x)
    x = x.reshape(x_shape)
    return x


def all_reduce_sync(x, *args, group=None, **kwargs):
    group = get_group(group)
    x = ft_c.all_reduce(x, *args, group=group, **kwargs)
    x = _maybe_wait(x)
    return x


def get_buffer(
    shape_or_tensor: Union[Tuple[int], torch.Tensor],
    *,
    repeats: int = 1,
    dim: int = 0,
    dtype: Optional[torch.dtype] = None,
    device: Optional[torch.device] = None,
    group=None,
) -> torch.Tensor:
    if repeats is None:
        repeats = get_world_size(group)

    if isinstance(shape_or_tensor, torch.Tensor):
        shape = shape_or_tensor.shape
        dtype = shape_or_tensor.dtype
        device = shape_or_tensor.device

    assert dtype is not None
    assert device is not None

    shape = list(shape)
    if repeats > 1:
        shape[dim] *= repeats

    buffer = torch.empty(shape, dtype=dtype, device=device)
    return buffer


def get_assigned_chunk(
    tensor: torch.Tensor,
    dim: int = 0,
    idx: Optional[int] = None,
    group=None,
) -> torch.Tensor:
    if idx is None:
        idx = get_rank(group)
    world_size = get_world_size(group)
    total_size = tensor.shape[dim]
    assert total_size % world_size == 0, f"tensor.shape[{dim}]={total_size} is not divisible by world_size={world_size}"
    return tensor.chunk(world_size, dim=dim)[idx]


def get_complete_tensor(
    tensor: torch.Tensor,
    *,
    dim: int = 0,
    group=None,
) -> torch.Tensor:
    tensor = tensor.transpose(0, dim).contiguous()
    output_tensor = all_gather_tensor_sync(tensor, gather_dim=0, group=group)
    output_tensor = output_tensor.transpose(0, dim)
    return output_tensor



import torch
from packaging import version

from torch.overrides import TorchFunctionMode

import para_attn


@torch.compiler.assume_constant_result
def torch_version_check(op, v):
    return getattr(version.parse(torch.__version__), op)(version.parse(v))


@torch.compiler.assume_constant_result
def get_force_dispatch_to_custom_ops():
    return para_attn.config.attention.force_dispatch_to_custom_ops


class BaseTorchFunctionMode(TorchFunctionMode):
    @torch.compiler.disable
    def __init__(self):
        super().__init__()


def base_handle_torch_function(func, types, args=(), kwargs=None):
    kwargs = {} if kwargs is None else kwargs

    if func is torch.Tensor.unflatten:
        # Reason: Unsupported: non-function or method super: <method 'unflatten' of 'torch._C.TensorBase' objects>
        return torch.unflatten(*args, **kwargs)

    return func(*args, **kwargs)



import torch
from diffusers import WanPipeline

# from diffusers.schedulers.scheduling_unipc_multistep import UniPCMultistepScheduler
from diffusers.utils import export_to_video

model_id = "Wan-AI/Wan2.1-T2V-1.3B-Diffusers"
# model_id = "Wan-AI/Wan2.1-T2V-14B-Diffusers"
pipe = WanPipeline.from_pretrained(model_id, torch_dtype=torch.bfloat16)

# flow shift should be 3.0 for 480p images, 5.0 for 720p images
# pipe.scheduler = UniPCMultistepScheduler.from_config(pipe.scheduler.config, flow_shift=3.0)
pipe.to("cuda")

from para_attn.first_block_cache.diffusers_adapters import apply_cache_on_pipe

apply_cache_on_pipe(pipe)

# Enable memory savings
# pipe.enable_model_cpu_offload()
# pipe.enable_vae_tiling()

# pipe.transformer = torch.compile(pipe.transformer, mode="max-autotune-no-cudagraphs")

video = pipe(
    prompt="An astronaut dancing vigorously on the moon with earth flying past in the background, hyperrealistic",
    negative_prompt="",
    height=480,
    width=832,
    num_frames=81,
    num_inference_steps=30,
).frames[0]

print("Saving video to wan.mp4")
export_to_video(video, "wan.mp4", fps=15)



import para_attn
import pytest

import torch


@pytest.hookimpl(optionalhook=True)
def pytest_html_report_title(report):
    """
    A pytest-html hook to add custom fields to the report.
    """
    # Customize the title of the html report (if used) to include version.
    device_count = torch.cuda.device_count()
    if device_count == 0:
        gpu_info = "no GPU"
    else:
        gpu_info = f"{device_count} {torch.cuda.get_device_name(0)} GPU(s)"
    para_attn_version = para_attn.__version__
    report.title = f"ParaAttention {para_attn_version} Tests with {gpu_info}"


@pytest.hookimpl(optionalhook=True)
def pytest_html_results_summary(prefix, summary, postfix):
    prefix.extend(["<p>System Environment Information</p>"])

    # Get system information
    env_info = torch.utils.collect_env.get_pretty_env_info()

    # Add system information to the summary
    prefix.extend([f"<pre>{env_info}</pre>"])



from typing import Optional, Tuple

import torch
import torch.backends.cuda as cuda_backend
import torch.nn.functional as F
from torch.backends.cuda import SDPAParams
from torch.nn.attention import SDPBackend

aten = torch.ops.aten


def cannot_use_attention_backend(*args, **kwargs):
    return False


can_use_flash_attention = getattr(cuda_backend, "can_use_flash_attention", cannot_use_attention_backend)
can_use_efficient_attention = getattr(cuda_backend, "can_use_efficient_attention", cannot_use_attention_backend)
can_use_cudnn_attention = getattr(cuda_backend, "can_use_cudnn_attention", cannot_use_attention_backend)

# torch.compile() support is only enabled for pytorch >= 2.4
# The reason for this is that we are using the new custom_op and register_fake
# APIs, which support inplace modification of inputs in the function itself
if torch.__version__ >= "2.4.0":
    _torch_custom_op_wrapper = torch.library.custom_op
    _torch_register_fake_wrapper = torch.library.register_fake
else:

    def noop_custom_op_wrapper(name, fn=None, /, *, mutates_args, device_types=None, schema=None):
        def wrap(func):
            return func

        if fn is None:
            return wrap
        return fn

    def noop_register_fake_wrapper(op, fn=None, /, *, lib=None, _stacklevel=1):
        def wrap(func):
            return func

        if fn is None:
            return wrap
        return fn

    _torch_custom_op_wrapper = noop_custom_op_wrapper
    _torch_register_fake_wrapper = noop_register_fake_wrapper


def flash_attention_forward_with_lse(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    attn_mask: Optional[torch.Tensor] = None,
    dropout_p: float = 0.0,
    is_causal: bool = False,
    *,
    scale: Optional[float] = None,
) -> Tuple[torch.Tensor, torch.Tensor]:
    return aten._scaled_dot_product_flash_attention(
        query,
        key,
        value,
        dropout_p=dropout_p,
        is_causal=is_causal,
        scale=scale,
    )[:2]


def efficient_attention_forward_with_lse(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    attn_mask: Optional[torch.Tensor] = None,
    dropout_p: float = 0.0,
    is_causal: bool = False,
    *,
    scale: Optional[float] = None,
) -> Tuple[torch.Tensor, torch.Tensor]:
    return aten._scaled_dot_product_efficient_attention(
        query,
        key,
        value,
        attn_bias=attn_mask,
        compute_log_sumexp=True,
        dropout_p=dropout_p,
        is_causal=is_causal,
        scale=scale,
    )[:2]


def cudnn_attention_forward_with_lse(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    attn_mask: Optional[torch.Tensor] = None,
    dropout_p: float = 0.0,
    is_causal: bool = False,
    *,
    scale: Optional[float] = None,
) -> Tuple[torch.Tensor, torch.Tensor]:
    return aten._scaled_dot_product_cudnn_attention(
        query,
        key,
        value,
        attn_bias=attn_mask,
        compute_log_sumexp=True,
        dropout_p=dropout_p,
        is_causal=is_causal,
        scale=scale,
    )[:2]


def _attention_forward_with_lse(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    attn_mask: Optional[torch.Tensor] = None,
    dropout_p: float = 0.0,
    is_causal: bool = False,
    *,
    scale: Optional[float] = None,
) -> Tuple[torch.Tensor, torch.Tensor]:
    device = query.device
    assert device.type == "cuda", "Must be CUDA tensors"
    assert torch.version.hip is None, "HIP is not supported"
    params_init_args = [
        query,
        key,
        value,
        attn_mask,
        dropout_p,
        is_causal,
    ]
    if hasattr(SDPAParams, "enable_gqa"):
        params_init_args.append(False)
    params = SDPAParams(*params_init_args)
    dispatches = [
        ["FLASH_ATTENTION", can_use_flash_attention, flash_attention_forward_with_lse],
        ["EFFICIENT_ATTENTION", can_use_efficient_attention, efficient_attention_forward_with_lse],
        ["CUDNN_ATTENTION", can_use_cudnn_attention, cudnn_attention_forward_with_lse],
    ]
    if hasattr(torch._C, "_get_sdp_priority_order"):
        priority_order = torch._C._get_sdp_priority_order()
        priority_order = [str(SDPBackend(i)).split(".")[-1] for i in priority_order]
    else:
        device_capability = torch.cuda.get_device_capability(device.index)
        if device_capability >= (9, 0):
            priority_order = ["CUDNN_ATTENTION", "FLASH_ATTENTION", "EFFICIENT_ATTENTION"]
        else:
            priority_order = ["FLASH_ATTENTION", "CUDNN_ATTENTION", "EFFICIENT_ATTENTION"]
    dispatches = [dispatch for dispatch in dispatches if dispatch[0] in priority_order]
    dispatches = sorted(dispatches, key=lambda x: priority_order.index(x[0]))
    for dispatch in dispatches:
        if dispatch[1](params):
            return dispatch[2](
                query, key, value, attn_mask=attn_mask, dropout_p=dropout_p, is_causal=is_causal, scale=scale
            )
    raise NotImplementedError


@_torch_custom_op_wrapper("para_attn::attention_forward_with_lse", mutates_args=(), device_types=("cpu", "cuda"))
def attention_forward_with_lse(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    attn_mask: Optional[torch.Tensor] = None,
    dropout_p: float = 0.0,
    is_causal: bool = False,
    *,
    scale: Optional[float] = None,
) -> Tuple[torch.Tensor, torch.Tensor]:
    return _attention_forward_with_lse(
        query,
        key,
        value,
        attn_mask=attn_mask,
        dropout_p=dropout_p,
        is_causal=is_causal,
        scale=scale,
    )


@_torch_register_fake_wrapper("para_attn::attention_forward_with_lse")
def _(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    attn_mask: Optional[torch.Tensor] = None,
    dropout_p: float = 0.0,
    is_causal: bool = False,
    *,
    scale: Optional[float] = None,
) -> Tuple[torch.Tensor, torch.Tensor]:
    return _attention_forward_with_lse(
        query,
        key,
        value,
        attn_mask=attn_mask,
        dropout_p=dropout_p,
        is_causal=is_causal,
        scale=scale,
    )


@_torch_custom_op_wrapper("para_attn::attention_forward", mutates_args=(), device_types=("cpu", "cuda"))
def attention_forward(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    attn_mask: Optional[torch.Tensor] = None,
    dropout_p: float = 0.0,
    is_causal: bool = False,
    *,
    scale: Optional[float] = None,
) -> torch.Tensor:
    return F.scaled_dot_product_attention(
        query,
        key,
        value,
        attn_mask=attn_mask,
        dropout_p=dropout_p,
        is_causal=is_causal,
        scale=scale,
    )


@_torch_register_fake_wrapper("para_attn::attention_forward")
def _(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    attn_mask: Optional[torch.Tensor] = None,
    dropout_p: float = 0.0,
    is_causal: bool = False,
    *,
    scale: Optional[float] = None,
) -> torch.Tensor:
    return F.scaled_dot_product_attention(
        query,
        key,
        value,
        attn_mask=attn_mask,
        dropout_p=dropout_p,
        is_causal=is_causal,
        scale=scale,
    )


def _attention_forward_sparse_kv(
    query,
    key,
    value,
    attn_mask=None,
    dropout_p=0.0,
    is_causal=False,
    *,
    scale=None,
):
    if attn_mask is None:
        return F.scaled_dot_product_attention(
            query,
            key,
            value,
            attn_mask=attn_mask,
            dropout_p=dropout_p,
            is_causal=is_causal,
            scale=scale,
        )

    assert attn_mask.dtype == torch.bool, "attn_mask must be a boolean tensor"
    assert not is_causal, "is_causal is not supported with sparse kv"

    s_kv = key.shape[-2]
    indices = torch.arange(s_kv, device=key.device)

    outputs = []
    for q, k, v, m in zip(query, key, value, attn_mask):
        q = q.unsqueeze(0)
        k = k.unsqueeze(0)
        v = v.unsqueeze(0)
        while m.ndim > 1:
            m = m[0]
        indices_ = indices[m]
        k = k[..., indices_, :]
        v = v[..., indices_, :]
        outputs.append(
            F.scaled_dot_product_attention(
                q,
                k,
                v,
                dropout_p=dropout_p,
                is_causal=is_causal,
                scale=scale,
            )
        )
        del q, k, v, m, indices_

    if len(outputs) == 1:
        return outputs[0]
    return torch.cat(outputs, dim=0)


@_torch_custom_op_wrapper("para_attn::attention_forward_sparse_kv", mutates_args=(), device_types=("cpu", "cuda"))
def attention_forward_sparse_kv(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    attn_mask: Optional[torch.Tensor] = None,
    dropout_p: float = 0.0,
    is_causal: bool = False,
    *,
    scale: Optional[float] = None,
) -> torch.Tensor:
    return _attention_forward_sparse_kv(
        query,
        key,
        value,
        attn_mask=attn_mask,
        dropout_p=dropout_p,
        is_causal=is_causal,
        scale=scale,
    )


@_torch_register_fake_wrapper("para_attn::attention_forward_sparse_kv")
def _(
    query: torch.Tensor,
    key: torch.Tensor,
    value: torch.Tensor,
    attn_mask: Optional[torch.Tensor] = None,
    dropout_p: float = 0.0,
    is_causal: bool = False,
    *,
    scale: Optional[float] = None,
) -> torch.Tensor:
    return _attention_forward_sparse_kv(
        query,
        key,
        value,
        # attn_mask=attn_mask,
        dropout_p=dropout_p,
        is_causal=is_causal,
        scale=scale,
    )



import torch
from diffusers import FluxPipeline

pipe = FluxPipeline.from_pretrained(
    "black-forest-labs/FLUX.1-dev",
    torch_dtype=torch.bfloat16,
).to("cuda")

from para_attn.first_block_cache.diffusers_adapters import apply_cache_on_pipe

apply_cache_on_pipe(pipe)

# pipe.enable_model_cpu_offload()

# pipe.transformer = torch.compile(pipe.transformer, mode="max-autotune-no-cudagraphs")

image = pipe(
    "A cat holding a sign that says hello world",
    num_inference_steps=28,
).images[0]

print("Saving image to flux.png")
image.save("flux.png")



import contextlib
import os
import socket
import threading
import time
import traceback

import torch.distributed as dist
import torch.multiprocessing as mp


class ExceptionOutput(Exception):
    pass


class MPDistRunner:
    def __init__(self, *, rank=None, persist_attrs=None, temp_attrs=None):
        self.rank = rank
        self.persist_attrs = persist_attrs or {}
        self.temp_attrs = temp_attrs or {}

        self.array_size = None
        self.shared_array = None
        self.input_queue = None
        self.output_queue = None
        self.status = None
        self.lock = None

        self.processes = []

    def __del__(self):
        self.terminate()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.terminate()

    @property
    def world_size(self):
        raise NotImplementedError

    @property
    def start_method(self):
        return "spawn"

    def clone(self, *, rank=None):
        return self.__class__(rank=rank, persist_attrs={**self.persist_attrs})

    def init_env(self, *, master_addr=None, master_port=None):
        if master_addr is None:
            if "MASTER_ADDR" not in os.environ:
                master_addr = "127.0.0.1"
                os.environ["MASTER_ADDR"] = master_addr
        else:
            os.environ["MASTER_ADDR"] = master_addr
        if master_port is None:
            if "MASTER_PORT" not in os.environ:
                master_port = find_free_port()
                os.environ["MASTER_PORT"] = str(master_port)
        else:
            os.environ["MASTER_PORT"] = str(master_port)

    def start(self, args=(), kwargs=None, *, timeout=None):
        if self.processes:
            raise RuntimeError("Processes are already started")

        if kwargs is None:
            kwargs = {}

        self.init_env()

        world_size = self.world_size
        if world_size <= 0:
            raise ValueError(f"World size {world_size} is invalid")

        start_method = self.start_method
        mp_ = mp.get_context(start_method)

        barrier = mp_.Barrier(world_size)
        input_queues = [mp_.JoinableQueue(maxsize=1) for _ in range(world_size)]
        output_queue = mp_.Queue(maxsize=1)
        exception_queues = [mp_.Queue(maxsize=1) for _ in range(world_size)]
        status = mp_.Value("i", 0, lock=True)
        processes = []

        try:
            for rank in range(world_size):
                process = mp_.Process(
                    target=self.worker,
                    args=(
                        self.clone(rank=rank),
                        args,
                        kwargs,
                        barrier,
                        input_queues[rank],
                        output_queue if rank == 0 else None,
                        exception_queues[rank],
                        status if rank == 0 else None,
                    ),
                    daemon=True,
                )
                process.start()
                processes.append(process)

            exceptions = []
            begin_time = time.time()
            while True:
                if timeout is not None and time.time() - begin_time >= timeout:
                    raise RuntimeError("Timeout during initialization")
                for rank, (process, exception_queue) in enumerate(zip(processes, exception_queues)):
                    if process.is_alive():
                        if exception_queue.empty():
                            continue
                        else:
                            exception = exception_queue.get()
                            if exception is None:
                                exceptions.append(None)
                            else:
                                exceptions.append((rank, exception))
                    else:
                        if exception_queue.empty():
                            exceptions.append((rank, RuntimeError(f"Process {rank} is not alive")))
                        else:
                            exception = exception_queue.get()
                            if exception is None:
                                exceptions.append(
                                    (rank, RuntimeError(f"Process {rank} is not alive after initialization"))
                                )
                            else:
                                exceptions.append((rank, exception))
                if len(exceptions) == world_size or any(e is not None for e in exceptions):
                    break
                time.sleep(0.0)
            exceptions = [e for e in exceptions if e is not None]
            if exceptions:
                msg = "\n".join(f"Rank {rank}: {exception}" for rank, exception in exceptions)
                raise RuntimeError(f"Exceptions occurred:\n{msg}")
        except Exception:
            self.terminate()
            raise

        self.input_queues = input_queues
        self.output_queue = output_queue
        self.exception_queues = exception_queues
        self.status = status
        self.lock = threading.RLock()
        self.processes = processes

        return self

    def terminate(self):
        self.input_queues = None
        self.output_queue = None
        self.exception_queues = None
        self.status = None

        for process in self.processes:
            if process.is_alive():
                process.terminate()
        self.processes = []

    def restart(self, *args, **kwargs):
        self.terminate()
        self.start(*args, **kwargs)

    def is_alive(self):
        return self.processes and all(process.is_alive() for process in self.processes)

    def is_idle(self):
        if not self.is_alive():
            return False

        with self.status.get_lock():
            return self.status.value == 0

    def is_almost_idle(self):
        if not self.is_alive():
            return False

        with self.status.get_lock():
            return self.status.value in (0, 2)

    def __call__(self, args=(), kwargs=None, *, timeout=None):
        if kwargs is None:
            kwargs = {}

        with self.lock:
            if not self.processes:
                raise RuntimeError("Processes are not started")
            for rank, process in enumerate(self.processes):
                if not process.is_alive():
                    raise RuntimeError(f"Process {rank} is not alive")

            if timeout is not None:
                begin_time = time.time()
            for input_queue in self.input_queues:
                input_queue.put((args, kwargs))
            for input_queue in self.input_queues:
                input_queue.join()
            if timeout is not None:
                end_time = time.time()
                duration = end_time - begin_time
                timeout = max(0.0, timeout - duration)
                if timeout == 0:
                    raise RuntimeError("Timeout during processing")
            begin_time = time.time()
            while True:
                if timeout is not None and time.time() - begin_time >= timeout:
                    raise RuntimeError("Timeout during initialization")
                if self.output_queue.empty():
                    for rank, process in enumerate(self.processes):
                        if not process.is_alive():
                            raise RuntimeError(f"Process {rank} is not alive")
                    time.sleep(0.0)
                    continue
                output = self.output_queue.get()
                exceptions = []
                for rank, exception_queue in enumerate(self.exception_queues):
                    if (rank == 0 and isinstance(output, ExceptionOutput)) or not exception_queue.empty():
                        exceptions.append((rank, exception_queue.get()))
                if exceptions:
                    msg = "\n".join(f"Rank {rank}: {exception}" for rank, exception in exceptions)
                    raise RuntimeError(f"Exceptions occurred:\n{msg}")
                if isinstance(output, ExceptionOutput):
                    raise RuntimeError("Exception occurred")
                return output

    def init_process_group(self):
        if dist.is_initialized():
            return
        dist.init_process_group(world_size=self.world_size, rank=self.rank)

    def destroy_process_group(self):
        if not dist.is_initialized():
            return
        dist.destroy_process_group()

    def init_processor(self):
        pass

    def process_task(self, *args, **kwargs):
        raise NotImplementedError

    @classmethod
    def worker(
        cls,
        runner,
        args,
        kwargs,
        barrier,
        input_queue,
        output_queue,
        exception_queue,
        status,
    ):
        runner.init_process_group()

        try:
            try:
                runner.init_processor(*args, **kwargs)
            except Exception as e:
                exception = RuntimeError(f"Failed to initialize processor: {e}\n{traceback.format_exc()}")
                if exception_queue is not None:
                    exception_queue.put(exception)
                # raise
                while True:
                    time.sleep(0.0)

            if exception_queue is not None:
                exception_queue.put(None)

            while True:
                input_args, input_kwargs = input_queue.get()

                if status is not None:
                    with status.get_lock():
                        status.value = 1

                output = None
                exception = None

                if output_queue is not None:
                    while not output_queue.empty():
                        output_queue.get()
                if exception_queue is not None:
                    while not exception_queue.empty():
                        exception_queue.get()
                input_queue.task_done()

                if exception is None:
                    try:
                        output = runner.process_task(*input_args, **input_kwargs)
                    except Exception as e:
                        exception = RuntimeError(f"Failed to process task: {e}\n{traceback.format_exc()}")

                if status is not None:
                    with status.get_lock():
                        status.value = 2

                if exception_queue is not None:
                    if exception is not None:
                        exception_queue.put(exception)

                if output_queue is not None:
                    if exception is None:
                        output_queue.put(output)
                    else:
                        output_queue.put(ExceptionOutput())

                barrier.wait()

                if status is not None:
                    with status.get_lock():
                        status.value = 0
        finally:
            runner.destroy_process_group()


def find_free_port():
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]



import importlib


def parallelize_vae(vae, *args, **kwargs) -> None:
    vae_cls_name = vae.__class__.__name__
    if False:
        pass
    elif vae_cls_name == "AutoencoderKL":
        adapter_name = "autoencoder_kl"
    elif vae_cls_name == "AutoencoderKLHunyuanVideo":
        adapter_name = "autoencoder_kl_hunyuan_video"
    elif vae_cls_name == "AutoencoderKLWan":
        adapter_name = "autoencoder_kl_wan"
    else:
        raise ValueError(f"Unknown vae class name: {vae_cls_name}")

    adapter_module = importlib.import_module(f".{adapter_name}", __package__)
    parallelize_vae_fn = getattr(adapter_module, "parallelize_vae")
    return parallelize_vae_fn(vae, *args, **kwargs)



import pytest

import torch

if not torch.cuda.is_available():
    pytest.skip("CUDA is not available", allow_module_level=True)

import torch.distributed as dist
import torch.nn.functional as F
from para_attn import para_attn_interface
from para_attn.distributed.mp_runner import MPDistRunner


class ParallelAttnRunner(MPDistRunner):
    @property
    def world_size(self):
        return torch.cuda.device_count()

    @property
    def attn_func(self):
        return None

    def attn_mode(self, device):
        return None

    def _test_attn_func_task(self, dtype, device, B, H, S_Q, S_KV, D, is_causal, compile):
        func = self.attn_func
        if func is None:
            return

        if compile:
            func = torch.compile(func)

        if is_causal and S_Q != S_KV:
            return

        with torch.no_grad(), torch.cuda.device(self.rank):
            torch.manual_seed(0)

            query = torch.randn(B, H, S_Q, D, dtype=dtype, device=device)
            key = torch.randn(B, H, S_KV, D, dtype=dtype, device=device)
            value = torch.randn(B, H, S_KV, D, dtype=dtype, device=device)
            attn_mask = None
            dropout_p = 0.0
            is_causal = is_causal

            query_slice = query.chunk(self.world_size, dim=-2)[self.rank]
            key_slice = key.chunk(self.world_size, dim=-2)[self.rank]
            value_slice = value.chunk(self.world_size, dim=-2)[self.rank]

            for _ in range(2 if compile else 1):
                out_slice = func(
                    query_slice,
                    key_slice,
                    value_slice,
                    attn_mask=attn_mask,
                    dropout_p=dropout_p,
                    is_causal=is_causal,
                )

            out_slice_ref = F.scaled_dot_product_attention(
                query,
                key,
                value,
                attn_mask=attn_mask,
                dropout_p=dropout_p,
                is_causal=is_causal,
            ).chunk(self.world_size, dim=-2)[self.rank]

            torch.testing.assert_close(out_slice, out_slice_ref, rtol=1e-5, atol=1e-3 * self.world_size)

    def _test_attn_mode_task(self, dtype, device, B, H, S_Q, S_KV, D, is_causal, compile):
        attn_mode = self.attn_mode(device)
        if attn_mode is None:
            return

        if is_causal and S_Q != S_KV:
            return

        with torch.no_grad(), torch.cuda.device(self.rank):
            torch.manual_seed(0)

            query = torch.randn(B, H, S_Q, D, dtype=dtype, device=device)
            key = torch.randn(B, H, S_KV, D, dtype=dtype, device=device)
            value = torch.randn(B, H, S_KV, D, dtype=dtype, device=device)
            attn_mask = None
            dropout_p = 0.0
            is_causal = is_causal

            query_slice = query.chunk(self.world_size, dim=-2)[self.rank]
            key_slice = key.chunk(self.world_size, dim=-2)[self.rank]
            value_slice = value.chunk(self.world_size, dim=-2)[self.rank]

            def func(*args, **kwargs):
                return F.scaled_dot_product_attention(*args, **kwargs)

            if compile:
                func = torch.compile(func, fullgraph=True)

            for _ in range(2 if compile else 1):
                with attn_mode:
                    out_slice = func(
                        query_slice,
                        key_slice,
                        value_slice,
                        attn_mask=attn_mask,
                        dropout_p=dropout_p,
                        is_causal=is_causal,
                    )

            out_slice_ref = F.scaled_dot_product_attention(
                query,
                key,
                value,
                attn_mask=attn_mask,
                dropout_p=dropout_p,
                is_causal=is_causal,
            ).chunk(self.world_size, dim=-2)[self.rank]

            torch.testing.assert_close(out_slice, out_slice_ref, rtol=1e-5, atol=1e-3 * self.world_size)

    def process_task(self, name, *args, **kwargs):
        return getattr(self, name)(*args, **kwargs)


class _TestParallelAttn:
    class Runner(ParallelAttnRunner):
        pass

    @pytest.mark.parametrize("dtype", [torch.float16, torch.bfloat16])
    @pytest.mark.parametrize("device", ["cuda"])
    @pytest.mark.parametrize(
        "B,H,S_Q,S_KV,D",
        [
            [2, 24, 4096, 4096, 64],
            [2, 24, 4096, 1024, 64],
        ],
    )
    @pytest.mark.parametrize("is_causal", [False, True])
    @pytest.mark.parametrize("compile", [False, True])
    def test_attn_func(self, dtype, device, B, H, S_Q, S_KV, D, is_causal, compile):
        with self.Runner().start() as runner:
            runner(
                ("_test_attn_func_task", dtype, device, B, H, S_Q, S_KV, D, is_causal, compile),
            )

    @pytest.mark.parametrize("dtype", [torch.float16, torch.bfloat16])
    @pytest.mark.parametrize("device", ["cuda"])
    @pytest.mark.parametrize(
        "B,H,S_Q,S_KV,D",
        [
            [2, 24, 4096, 4096, 64],
            [2, 24, 4096, 1024, 64],
        ],
    )
    @pytest.mark.parametrize("is_causal", [False, True])
    @pytest.mark.parametrize("compile", [False, True])
    def test_attn_mode(self, dtype, device, B, H, S_Q, S_KV, D, is_causal, compile):
        with self.Runner().start() as runner:
            runner(
                ("_test_attn_mode_task", dtype, device, B, H, S_Q, S_KV, D, is_causal, compile),
            )


class TestRingAttn(_TestParallelAttn):
    class Runner(ParallelAttnRunner):
        @property
        def attn_func(self):
            return para_attn_interface.ring_attn_func

        def attn_mode(self, device):
            return para_attn_interface.RingAttnMode()


class TestUlyssesAttn(_TestParallelAttn):
    class Runner(ParallelAttnRunner):
        @property
        def attn_func(self):
            return para_attn_interface.ulysses_attn_func

        def attn_mode(self, device):
            return para_attn_interface.UlyssesAttnMode()


class TestUnifiedAttn(_TestParallelAttn):
    class Runner(ParallelAttnRunner):
        def attn_mode(self, device):
            world_size = self.world_size
            if world_size % 2 == 0:
                mesh_shape = (2, world_size // 2)
            else:
                mesh_shape = (1, world_size)
            mesh = dist.init_device_mesh(device, mesh_shape, mesh_dim_names=("ring", "ulysses"))
            return para_attn_interface.UnifiedAttnMode(mesh)



import functools
import unittest
from typing import Any, Dict, Optional, Union

import torch
from diffusers import DiffusionPipeline, HunyuanVideoTransformer3DModel
from diffusers.models.modeling_outputs import Transformer2DModelOutput
from diffusers.utils import logging, scale_lora_layers, unscale_lora_layers, USE_PEFT_BACKEND

from para_attn.first_block_cache import utils
from para_attn.para_attn_interface import SparseKVAttnMode

logger = logging.get_logger(__name__)  # pylint: disable=invalid-name


def apply_cache_on_transformer(
    transformer: HunyuanVideoTransformer3DModel,
):
    if getattr(transformer, "_is_cached", False):
        return transformer

    cached_transformer_blocks = torch.nn.ModuleList(
        [
            utils.CachedTransformerBlocks(
                transformer.transformer_blocks + transformer.single_transformer_blocks,
                transformer=transformer,
            )
        ]
    )
    dummy_single_transformer_blocks = torch.nn.ModuleList()

    original_forward = transformer.forward

    @functools.wraps(transformer.__class__.forward)
    def new_forward(
        self,
        hidden_states: torch.Tensor,
        timestep: torch.LongTensor,
        encoder_hidden_states: torch.Tensor,
        encoder_attention_mask: torch.Tensor,
        pooled_projections: torch.Tensor,
        guidance: torch.Tensor = None,
        attention_kwargs: Optional[Dict[str, Any]] = None,
        return_dict: bool = True,
        **kwargs,
    ) -> Union[torch.Tensor, Dict[str, torch.Tensor]]:
        with unittest.mock.patch.object(
            self,
            "transformer_blocks",
            cached_transformer_blocks,
        ), unittest.mock.patch.object(
            self,
            "single_transformer_blocks",
            dummy_single_transformer_blocks,
        ):
            if getattr(self, "_is_parallelized", False):
                return original_forward(
                    hidden_states,
                    timestep,
                    encoder_hidden_states,
                    encoder_attention_mask,
                    pooled_projections,
                    guidance=guidance,
                    attention_kwargs=attention_kwargs,
                    return_dict=return_dict,
                    **kwargs,
                )
            else:
                if attention_kwargs is not None:
                    attention_kwargs = attention_kwargs.copy()
                    lora_scale = attention_kwargs.pop("scale", 1.0)
                else:
                    lora_scale = 1.0

                if USE_PEFT_BACKEND:
                    # weight the lora layers by setting `lora_scale` for each PEFT layer
                    scale_lora_layers(self, lora_scale)
                else:
                    if attention_kwargs is not None and attention_kwargs.get("scale", None) is not None:
                        logger.warning(
                            "Passing `scale` via `attention_kwargs` when not using the PEFT backend is ineffective."
                        )

                batch_size, num_channels, num_frames, height, width = hidden_states.shape
                p, p_t = self.config.patch_size, self.config.patch_size_t
                post_patch_num_frames = num_frames // p_t
                post_patch_height = height // p
                post_patch_width = width // p

                # 1. RoPE
                image_rotary_emb = self.rope(hidden_states)

                # 2. Conditional embeddings
                temb = self.time_text_embed(timestep, guidance, pooled_projections)
                hidden_states = self.x_embedder(hidden_states)
                encoder_hidden_states = self.context_embedder(encoder_hidden_states, timestep, encoder_attention_mask)

                # 3. Attention mask preparation
                latent_sequence_length = hidden_states.shape[1]
                latent_attention_mask = torch.ones(
                    batch_size, 1, latent_sequence_length, device=hidden_states.device, dtype=torch.bool
                )  # [B, 1, N]
                attention_mask = torch.cat(
                    [latent_attention_mask, encoder_attention_mask.unsqueeze(1).to(torch.bool)], dim=-1
                )  # [B, 1, N + M]

                with SparseKVAttnMode():
                    # 4. Transformer blocks
                    hidden_states, encoder_hidden_states = self.call_transformer_blocks(
                        hidden_states, encoder_hidden_states, temb, attention_mask, image_rotary_emb
                    )

                # 5. Output projection
                hidden_states = self.norm_out(hidden_states, temb)
                hidden_states = self.proj_out(hidden_states)

                hidden_states = hidden_states.reshape(
                    batch_size, post_patch_num_frames, post_patch_height, post_patch_width, -1, p_t, p, p
                )
                hidden_states = hidden_states.permute(0, 4, 1, 5, 2, 6, 3, 7)
                hidden_states = hidden_states.flatten(6, 7).flatten(4, 5).flatten(2, 3)

                hidden_states = hidden_states.to(timestep.dtype)

                if USE_PEFT_BACKEND:
                    # remove `lora_scale` from each PEFT layer
                    unscale_lora_layers(self, lora_scale)

                if not return_dict:
                    return (hidden_states,)

                return Transformer2DModelOutput(sample=hidden_states)

    transformer.forward = new_forward.__get__(transformer)

    def call_transformer_blocks(self, hidden_states, encoder_hidden_states, *args, **kwargs):
        if torch.is_grad_enabled() and self.gradient_checkpointing:

            def create_custom_forward(module, return_dict=None):
                def custom_forward(*inputs):
                    if return_dict is not None:
                        return module(*inputs, return_dict=return_dict)
                    else:
                        return module(*inputs)

                return custom_forward

            ckpt_kwargs: Dict[str, Any] = {"use_reentrant": False}

            for block in self.transformer_blocks:
                hidden_states, encoder_hidden_states = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(block),
                    hidden_states,
                    encoder_hidden_states,
                    *args,
                    **kwargs,
                    **ckpt_kwargs,
                )

            for block in self.single_transformer_blocks:
                hidden_states, encoder_hidden_states = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(block),
                    hidden_states,
                    encoder_hidden_states,
                    *args,
                    **kwargs,
                    **ckpt_kwargs,
                )

        else:
            for block in self.transformer_blocks:
                hidden_states, encoder_hidden_states = block(hidden_states, encoder_hidden_states, *args, **kwargs)

            for block in self.single_transformer_blocks:
                hidden_states, encoder_hidden_states = block(hidden_states, encoder_hidden_states, *args, **kwargs)

        return hidden_states, encoder_hidden_states

    transformer.call_transformer_blocks = call_transformer_blocks.__get__(transformer)

    transformer._is_cached = True

    return transformer


def apply_cache_on_pipe(
    pipe: DiffusionPipeline,
    *,
    shallow_patch: bool = False,
    residual_diff_threshold=0.06,
    downsample_factor=1,
    **kwargs,
):
    if not getattr(pipe, "_is_cached", False):
        original_call = pipe.__class__.__call__

        @functools.wraps(original_call)
        def new_call(self, *args, **kwargs):
            with utils.cache_context(
                utils.create_cache_context(
                    residual_diff_threshold=residual_diff_threshold,
                    downsample_factor=downsample_factor,
                )
            ):
                return original_call(self, *args, **kwargs)

        pipe.__class__.__call__ = new_call
        pipe.__class__._is_cached = True

    if not shallow_patch:
        apply_cache_on_transformer(pipe.transformer, **kwargs)

    return pipe



import os  # noqa: C101
import sys

# import torch


_save_config_ignore = {
    # workaround: "Can't pickle <function ...>"
}


class attention:
    # https://github.com/pytorch/pytorch/blob/8dd380803c0e25786cba12801088c420a2ca071b/aten/src/ATen/native/transformers/attention.cpp#L574
    # https://github.com/pytorch/pytorch/blob/8dd380803c0e25786cba12801088c420a2ca071b/torch/_inductor/lowering.py#L2450
    force_dispatch_to_custom_ops = os.getenv("PARA_ATTN_FORCE_DISPATCH_TO_CUSTOM_OPS") == "1"

    allow_reduced_precision_compute = os.getenv("PARA_ATTN_ALLOW_REDUCED_PRECISION_COMPUTE") == "1"

    # fast_math = os.getenv("PARA_ATTN_FAST_MATH") == "1"


try:
    from torch.utils._config_module import install_config_module
except ImportError:
    # torch<2.2.0
    from torch._dynamo.config_utils import install_config_module

# adds patch, save_config, etc
install_config_module(sys.modules[__name__])



import functools
import unittest

import torch
from diffusers import CogVideoXTransformer3DModel, DiffusionPipeline

from para_attn.first_block_cache import utils


def apply_cache_on_transformer(
    transformer: CogVideoXTransformer3DModel,
):
    if getattr(transformer, "_is_cached", False):
        return transformer

    cached_transformer_blocks = torch.nn.ModuleList(
        [
            utils.CachedTransformerBlocks(
                transformer.transformer_blocks,
                transformer=transformer,
            )
        ]
    )

    original_forward = transformer.forward

    @functools.wraps(transformer.__class__.forward)
    def new_forward(
        self,
        *args,
        **kwargs,
    ):
        with unittest.mock.patch.object(
            self,
            "transformer_blocks",
            cached_transformer_blocks,
        ):
            return original_forward(
                *args,
                **kwargs,
            )

    transformer.forward = new_forward.__get__(transformer)

    transformer._is_cached = True

    return transformer


def apply_cache_on_pipe(
    pipe: DiffusionPipeline,
    *,
    shallow_patch: bool = False,
    residual_diff_threshold=0.04,
    downsample_factor=1,
    **kwargs,
):
    if not getattr(pipe, "_is_cached", False):
        original_call = pipe.__class__.__call__

        @functools.wraps(original_call)
        def new_call(self, *args, **kwargs):
            with utils.cache_context(
                utils.create_cache_context(
                    residual_diff_threshold=residual_diff_threshold,
                    downsample_factor=downsample_factor,
                )
            ):
                return original_call(self, *args, **kwargs)

        pipe.__class__.__call__ = new_call
        pipe.__class__._is_cached = True

    if not shallow_patch:
        apply_cache_on_transformer(pipe.transformer, **kwargs)

    return pipe



import contextlib
import dataclasses
from collections import defaultdict
from typing import Any, DefaultDict, Dict, List, Optional, Union

import torch

import para_attn.primitives as DP
from .taylorseer import TaylorSeer


@dataclasses.dataclass
class CacheContext:
    residual_diff_threshold: Union[torch.Tensor, float] = 0.0
    alter_residual_diff_threshold: Optional[Union[torch.Tensor, float]] = None

    downsample_factor: int = 1

    enable_alter_cache: bool = False
    num_inference_steps: int = -1
    warmup_steps: int = 0

    enable_taylorseer: bool = False
    taylorseer_kwargs: Dict[str, Any] = dataclasses.field(default_factory=dict)

    slg_layers: Optional[List[int]] = None
    slg_start: float = 0.0
    slg_end: float = 0.1

    taylorseer: Optional[TaylorSeer] = None
    alter_taylorseer: Optional[TaylorSeer] = None

    buffers: Dict[str, Any] = dataclasses.field(default_factory=dict)
    incremental_name_counters: DefaultDict[str, int] = dataclasses.field(default_factory=lambda: defaultdict(int))

    executed_steps: int = 0
    is_alter_cache: bool = True

    def __post_init__(self):
        if self.enable_taylorseer:
            self.taylorseer = TaylorSeer(**self.taylorseer_kwargs)
            if self.enable_alter_cache:
                self.alter_taylorseer = TaylorSeer(**self.taylorseer_kwargs)

    def get_incremental_name(self, name=None):
        if name is None:
            name = "default"
        idx = self.incremental_name_counters[name]
        self.incremental_name_counters[name] += 1
        return f"{name}_{idx}"

    def reset_incremental_names(self):
        self.incremental_name_counters.clear()

    def get_residual_diff_threshold(self):
        if self.enable_alter_cache and self.is_alter_cache and self.alter_residual_diff_threshold is not None:
            residual_diff_threshold = self.alter_residual_diff_threshold
        else:
            residual_diff_threshold = self.residual_diff_threshold
        if isinstance(residual_diff_threshold, torch.Tensor):
            residual_diff_threshold = residual_diff_threshold.item()
        return residual_diff_threshold

    def get_buffer(self, name):
        if self.enable_alter_cache and self.is_alter_cache:
            name = f"{name}_alter"
        return self.buffers.get(name)

    def set_buffer(self, name, buffer):
        if self.enable_alter_cache and self.is_alter_cache:
            name = f"{name}_alter"
        self.buffers[name] = buffer

    def remove_buffer(self, name):
        if self.enable_alter_cache and self.is_alter_cache:
            name = f"{name}_alter"
        if name in self.buffers:
            del self.buffers[name]

    def clear_buffers(self):
        self.buffers.clear()

    def mark_step_begin(self):
        if not self.enable_alter_cache:
            self.executed_steps += 1
        else:
            self.is_alter_cache = not self.is_alter_cache
            if not self.is_alter_cache:
                self.executed_steps += 1
        if self.enable_taylorseer:
            taylorseer = self.get_taylorseer()
            taylorseer.mark_step_begin()

    def get_taylorseer(self):
        if self.enable_alter_cache and self.is_alter_cache:
            return self.alter_taylorseer
        return self.taylorseer

    def is_slg_enabled(self):
        return self.slg_layers is not None

    def slg_should_skip_block(self, block_idx):
        if not self.enable_alter_cache or not self.is_alter_cache:
            return False
        if self.slg_layers is None:
            return False
        if self.slg_start <= 0.0 and self.slg_end >= 1.0:
            return False
        num_inference_steps = self.num_inference_steps
        assert num_inference_steps >= 0, "num_inference_steps must be non-negative"
        return (
            block_idx in self.slg_layers
            and num_inference_steps * self.slg_start <= self.get_current_step() < num_inference_steps * self.slg_end
        )

    def get_current_step(self):
        return self.executed_steps - 1

    def is_in_warmup(self):
        return self.get_current_step() < self.warmup_steps


@torch.compiler.disable
def get_residual_diff_threshold():
    cache_context = get_current_cache_context()
    assert cache_context is not None, "cache_context must be set before"
    return cache_context.get_residual_diff_threshold()


@torch.compiler.disable
def get_buffer(name):
    cache_context = get_current_cache_context()
    assert cache_context is not None, "cache_context must be set before"
    return cache_context.get_buffer(name)


@torch.compiler.disable
def set_buffer(name, buffer):
    cache_context = get_current_cache_context()
    assert cache_context is not None, "cache_context must be set before"
    cache_context.set_buffer(name, buffer)


@torch.compiler.disable
def remove_buffer(name):
    cache_context = get_current_cache_context()
    assert cache_context is not None, "cache_context must be set before"
    cache_context.remove_buffer(name)


@torch.compiler.disable
def mark_step_begin():
    cache_context = get_current_cache_context()
    assert cache_context is not None, "cache_context must be set before"
    cache_context.mark_step_begin()


@torch.compiler.disable
def is_taylorseer_enabled():
    cache_context = get_current_cache_context()
    assert cache_context is not None, "cache_context must be set before"
    return cache_context.enable_taylorseer


@torch.compiler.disable
def get_taylorseer():
    cache_context = get_current_cache_context()
    assert cache_context is not None, "cache_context must be set before"
    return cache_context.get_taylorseer()


@torch.compiler.disable
def is_slg_enabled():
    cache_context = get_current_cache_context()
    assert cache_context is not None, "cache_context must be set before"
    return cache_context.is_slg_enabled()


@torch.compiler.disable
def slg_should_skip_block(block_idx):
    cache_context = get_current_cache_context()
    assert cache_context is not None, "cache_context must be set before"
    return cache_context.slg_should_skip_block(block_idx)


@torch.compiler.disable
def is_in_warmup():
    cache_context = get_current_cache_context()
    assert cache_context is not None, "cache_context must be set before"
    return cache_context.is_in_warmup()


_current_cache_context = None


def create_cache_context(*args, **kwargs):
    return CacheContext(*args, **kwargs)


def get_current_cache_context():
    return _current_cache_context


def set_current_cache_context(cache_context=None):
    global _current_cache_context
    _current_cache_context = cache_context


@contextlib.contextmanager
def cache_context(cache_context):
    global _current_cache_context
    old_cache_context = _current_cache_context
    _current_cache_context = cache_context
    try:
        yield
    finally:
        _current_cache_context = old_cache_context


@torch.compiler.disable
def are_two_tensors_similar(t1, t2, *, threshold, parallelized=False):
    if threshold <= 0.0:
        return False

    if t1.shape != t2.shape:
        return False

    mean_diff = (t1 - t2).abs().mean()
    mean_t1 = t1.abs().mean()
    if parallelized:
        mean_diff = DP.all_reduce_sync(mean_diff, "avg")
        mean_t1 = DP.all_reduce_sync(mean_t1, "avg")
    diff = mean_diff / mean_t1
    return diff.item() < threshold


@torch.compiler.disable
def apply_prev_hidden_states_residual(hidden_states, encoder_hidden_states):
    if is_taylorseer_enabled():
        hidden_states_residual = get_hidden_states_residual()
        assert hidden_states_residual is not None, "hidden_states_residual must be set before"
        hidden_states = hidden_states_residual + hidden_states

        hidden_states = hidden_states.contiguous()
    else:
        hidden_states_residual = get_hidden_states_residual()
        assert hidden_states_residual is not None, "hidden_states_residual must be set before"
        hidden_states = hidden_states_residual + hidden_states

        encoder_hidden_states_residual = get_encoder_hidden_states_residual()
        assert encoder_hidden_states_residual is not None, "encoder_hidden_states_residual must be set before"
        encoder_hidden_states = encoder_hidden_states_residual + encoder_hidden_states

        hidden_states = hidden_states.contiguous()
        encoder_hidden_states = encoder_hidden_states.contiguous()

    return hidden_states, encoder_hidden_states


@torch.compiler.disable
def get_downsample_factor():
    cache_context = get_current_cache_context()
    assert cache_context is not None, "cache_context must be set before"
    return cache_context.downsample_factor


@torch.compiler.disable
def get_can_use_cache(first_hidden_states_residual, parallelized=False):
    if is_in_warmup():
        return False
    threshold = get_residual_diff_threshold()
    if threshold <= 0.0:
        return False
    downsample_factor = get_downsample_factor()
    if downsample_factor > 1:
        first_hidden_states_residual = first_hidden_states_residual[..., ::downsample_factor]
    prev_first_hidden_states_residual = get_first_hidden_states_residual()
    can_use_cache = prev_first_hidden_states_residual is not None and are_two_tensors_similar(
        prev_first_hidden_states_residual,
        first_hidden_states_residual,
        threshold=threshold,
        parallelized=parallelized,
    )
    return can_use_cache


@torch.compiler.disable
def set_first_hidden_states_residual(first_hidden_states_residual):
    downsample_factor = get_downsample_factor()
    if downsample_factor > 1:
        first_hidden_states_residual = first_hidden_states_residual[..., ::downsample_factor]
        first_hidden_states_residual = first_hidden_states_residual.contiguous()
    set_buffer("first_hidden_states_residual", first_hidden_states_residual)


@torch.compiler.disable
def get_first_hidden_states_residual():
    return get_buffer("first_hidden_states_residual")


@torch.compiler.disable
def set_hidden_states_residual(hidden_states_residual):
    if is_taylorseer_enabled():
        taylorseer = get_taylorseer()
        taylorseer.update(hidden_states_residual)
    else:
        set_buffer("hidden_states_residual", hidden_states_residual)


@torch.compiler.disable
def get_hidden_states_residual():
    if is_taylorseer_enabled():
        taylorseer = get_taylorseer()
        return taylorseer.approximate_value()
    else:
        return get_buffer("hidden_states_residual")


@torch.compiler.disable
def set_encoder_hidden_states_residual(encoder_hidden_states_residual):
    if is_taylorseer_enabled():
        return
    set_buffer("encoder_hidden_states_residual", encoder_hidden_states_residual)


@torch.compiler.disable
def get_encoder_hidden_states_residual():
    return get_buffer("encoder_hidden_states_residual")


class CachedTransformerBlocks(torch.nn.Module):
    def __init__(
        self,
        transformer_blocks,
        single_transformer_blocks=None,
        *,
        transformer=None,
        return_hidden_states_first=True,
        return_hidden_states_only=False,
    ):
        super().__init__()

        self.transformer = transformer
        self.transformer_blocks = transformer_blocks
        self.single_transformer_blocks = single_transformer_blocks
        self.return_hidden_states_first = return_hidden_states_first
        self.return_hidden_states_only = return_hidden_states_only

    def forward(self, hidden_states, encoder_hidden_states, *args, **kwargs):
        original_hidden_states = hidden_states
        first_transformer_block = self.transformer_blocks[0]
        hidden_states = first_transformer_block(hidden_states, encoder_hidden_states, *args, **kwargs)
        if not isinstance(hidden_states, torch.Tensor):
            hidden_states, encoder_hidden_states = hidden_states
            if not self.return_hidden_states_first:
                hidden_states, encoder_hidden_states = encoder_hidden_states, hidden_states
        first_hidden_states_residual = hidden_states - original_hidden_states
        del original_hidden_states

        mark_step_begin()
        can_use_cache = get_can_use_cache(
            first_hidden_states_residual,
            parallelized=self.transformer is not None and getattr(self.transformer, "_is_parallelized", False),
        )

        torch._dynamo.graph_break()
        if can_use_cache:
            del first_hidden_states_residual
            hidden_states, encoder_hidden_states = apply_prev_hidden_states_residual(
                hidden_states, encoder_hidden_states
            )
        else:
            set_first_hidden_states_residual(first_hidden_states_residual)
            del first_hidden_states_residual
            (
                hidden_states,
                encoder_hidden_states,
                hidden_states_residual,
                encoder_hidden_states_residual,
            ) = self.call_remaining_transformer_blocks(hidden_states, encoder_hidden_states, *args, **kwargs)
            set_hidden_states_residual(hidden_states_residual)
            set_encoder_hidden_states_residual(encoder_hidden_states_residual)
        torch._dynamo.graph_break()

        return (
            hidden_states
            if self.return_hidden_states_only
            else (
                (hidden_states, encoder_hidden_states)
                if self.return_hidden_states_first
                else (encoder_hidden_states, hidden_states)
            )
        )

    def call_remaining_transformer_blocks(self, hidden_states, encoder_hidden_states, *args, **kwargs):
        original_hidden_states = hidden_states
        original_encoder_hidden_states = encoder_hidden_states
        if not is_slg_enabled():
            for block in self.transformer_blocks[1:]:
                hidden_states = block(hidden_states, encoder_hidden_states, *args, **kwargs)
                if not isinstance(hidden_states, torch.Tensor):
                    hidden_states, encoder_hidden_states = hidden_states
                    if not self.return_hidden_states_first:
                        hidden_states, encoder_hidden_states = encoder_hidden_states, hidden_states
            if self.single_transformer_blocks is not None:
                hidden_states = torch.cat([encoder_hidden_states, hidden_states], dim=1)
                for block in self.single_transformer_blocks:
                    hidden_states = block(hidden_states, *args, **kwargs)
                encoder_hidden_states, hidden_states = hidden_states.split(
                    [encoder_hidden_states.shape[1], hidden_states.shape[1] - encoder_hidden_states.shape[1]], dim=1
                )
        else:
            for i, encoder_block in enumerate(self.transformer_blocks[1:]):
                if slg_should_skip_block(i + 1):
                    continue
                hidden_states = encoder_block(hidden_states, encoder_hidden_states, *args, **kwargs)
                if not isinstance(hidden_states, torch.Tensor):
                    hidden_states, encoder_hidden_states = hidden_states
                    if not self.return_hidden_states_first:
                        hidden_states, encoder_hidden_states = encoder_hidden_states, hidden_states
            if self.single_transformer_blocks is not None:
                hidden_states = torch.cat([encoder_hidden_states, hidden_states], dim=1)
                for i, block in enumerate(self.single_transformer_blocks):
                    if slg_should_skip_block(len(self.transformer_blocks) + i):
                        continue
                    hidden_states = block(hidden_states, *args, **kwargs)
                encoder_hidden_states, hidden_states = hidden_states.split(
                    [encoder_hidden_states.shape[1], hidden_states.shape[1] - encoder_hidden_states.shape[1]], dim=1
                )

        # hidden_states_shape = hidden_states.shape
        # encoder_hidden_states_shape = encoder_hidden_states.shape
        hidden_states = hidden_states.reshape(-1).contiguous().reshape(original_hidden_states.shape)
        encoder_hidden_states = (
            encoder_hidden_states.reshape(-1).contiguous().reshape(original_encoder_hidden_states.shape)
        )

        # hidden_states = hidden_states.contiguous()
        # encoder_hidden_states = encoder_hidden_states.contiguous()

        hidden_states_residual = hidden_states - original_hidden_states
        encoder_hidden_states_residual = encoder_hidden_states - original_encoder_hidden_states

        hidden_states_residual = hidden_states_residual.reshape(-1).contiguous().reshape(original_hidden_states.shape)
        encoder_hidden_states_residual = (
            encoder_hidden_states_residual.reshape(-1).contiguous().reshape(original_encoder_hidden_states.shape)
        )

        return hidden_states, encoder_hidden_states, hidden_states_residual, encoder_hidden_states_residual



import torch
from diffusers import CogVideoXPipeline
from diffusers.utils import export_to_video

pipe = CogVideoXPipeline.from_pretrained(
    "THUDM/CogVideoX-5b",
    torch_dtype=torch.bfloat16,
).to("cuda")

from para_attn.first_block_cache.diffusers_adapters import apply_cache_on_pipe

apply_cache_on_pipe(pipe)

# pipe.enable_model_cpu_offload()
# pipe.enable_sequential_cpu_offload()
pipe.vae.enable_slicing()
pipe.vae.enable_tiling()


# pipe.transformer = torch.compile(pipe.transformer, mode="max-autotune-no-cudagraphs")

prompt = "A panda, dressed in a small, red jacket and a tiny hat, sits on a wooden stool in a serene bamboo forest. The panda's fluffy paws strum a miniature acoustic guitar, producing soft, melodic tunes. Nearby, a few other pandas gather, watching curiously and some clapping in rhythm. Sunlight filters through the tall bamboo, casting a gentle glow on the scene. The panda's face is expressive, showing concentration and joy as it plays. The background includes a small, flowing stream and vibrant green foliage, enhancing the peaceful and magical atmosphere of this unique musical performance."
video = pipe(
    prompt=prompt,
    num_videos_per_prompt=1,
    num_inference_steps=50,
    num_frames=49,
    guidance_scale=6,
    # generator=torch.Generator(device=pipe.device).manual_seed(42),
).frames[0]

print("Saving video to cogvideox.mp4")
export_to_video(video, "cogvideox.mp4", fps=8)



import contextlib

import torch
import torch.distributed as dist
import torch.nn.functional as F

import para_attn
import para_attn.ops as para_attn_ops
import para_attn.primitives as DP
from para_attn.sparse_attn import (
    focus_attn_func,
    FocusAttnMode,
    sparse_kv_attn_func,
    SparseKVAttnMode,
    struct_sparse_attn_func,
    StructSparseAttnMode,
)
from para_attn.utils import (
    base_handle_torch_function,
    BaseTorchFunctionMode,
    get_force_dispatch_to_custom_ops,
    torch_version_check,
)

try:
    from torch.distributed.tensor.experimental._attention import _templated_ring_attention
except ImportError:
    _templated_ring_attention = None

try:
    import torch.distributed.tensor.experimental._attention as torch_ring_attention
except ImportError:
    torch_ring_attention = None

__all__ = [
    "UnifiedAttnMode",
    "RingAttnMode",
    "UlyssesAttnMode",
    "InBatchAttnMode",
    "SparseKVAttnMode",
    "StructSparseAttnMode",
    "FocusAttnMode",
    "ring_attn_func",
    "ulysses_attn_func",
    "in_batch_attn_func",
    "sparse_kv_attn_func",
    "struct_sparse_attn_func",
    "focus_attn_func",
]


def _sdpa_all_to_all_single(x, mesh):
    if x.requires_grad:
        x = DP.all_to_all_single_autograd_sync(x, output_split_sizes=None, input_split_sizes=None, group=mesh)
    else:
        x = DP.all_to_all_single_sync(x, output_split_sizes=None, input_split_sizes=None, group=mesh)
    return x


def _sdpa_input_all_to_all(x, mesh):
    world_size = DP.get_world_size(mesh)
    if world_size <= 1:
        return x

    assert x.ndim == 4, "x must have 4 dimensions, got {}".format(x.ndim)
    b, h, s, d = x.shape
    assert h % world_size == 0, "h must be divisible by world_size, got {} and {}".format(h, world_size)

    x = x.permute(1, 0, 2, 3).contiguous()
    x = _sdpa_all_to_all_single(x, mesh)
    x = x.reshape(world_size, h // world_size, b, -1, d).permute(2, 1, 0, 3, 4).reshape(b, h // world_size, -1, d)
    return x


def _sdpa_output_all_to_all(x, mesh):
    world_size = DP.get_world_size(mesh)
    if world_size <= 1:
        return x

    assert x.ndim == 4, "x must have 4 dimensions, got {}".format(x.ndim)
    b, h, s, d = x.shape
    assert s % world_size == 0, "s must be divisible by world_size, got {} and {}".format(s, world_size)

    x = x.permute(2, 0, 1, 3).contiguous()
    x = _sdpa_all_to_all_single(x, mesh)
    x = x.reshape(world_size, s // world_size, b, -1, d).permute(2, 0, 3, 1, 4).reshape(b, -1, s // world_size, d)
    return x


def ulysses_attn_func(
    query,
    key,
    value,
    attn_mask=None,
    dropout_p=0.0,
    is_causal=False,
    *,
    scale=None,
    mesh=None,
    attn_func=None,
):
    assert query.ndim == 4, "query must have 4 dimensions, got {}".format(query.ndim)
    assert key.ndim == 4, "key must have 4 dimensions, got {}".format(key.ndim)
    assert value.ndim == 4, "value must have 4 dimensions, got {}".format(value.ndim)

    if mesh is None:
        mesh = DP.get_group()

    query = _sdpa_input_all_to_all(query, mesh)
    key = _sdpa_input_all_to_all(key, mesh)
    value = _sdpa_input_all_to_all(value, mesh)

    if attn_func is None:
        attn_func = F.scaled_dot_product_attention

    out = attn_func(query, key, value, attn_mask=attn_mask, dropout_p=dropout_p, is_causal=is_causal, scale=scale)

    out = _sdpa_output_all_to_all(out, mesh)
    return out


@torch.compiler.assume_constant_result
def _setup_ring_attn_func_forward():
    _convert_to_f32 = None
    _cp_options_convert_to_f32 = None
    _cp_options_enable_load_balance = None
    _cp_options_rotate_method = None

    assert torch_ring_attention is not None, "RingAttnFunc requires a newer version of PyTorch"
    _convert_to_f32 = getattr(torch_ring_attention, "_convert_to_f32", None)
    if _convert_to_f32 is not None:
        torch_ring_attention._convert_to_f32 = not para_attn.config.attention.allow_reduced_precision_compute
    _cp_options = getattr(torch_ring_attention, "_cp_options", None)
    if _cp_options is not None:
        _cp_options_convert_to_f32 = getattr(_cp_options, "convert_to_f32", None)
        if _cp_options_convert_to_f32 is not None:
            _cp_options.convert_to_f32 = not para_attn.config.attention.allow_reduced_precision_compute
        _cp_options_enable_load_balance = getattr(_cp_options, "enable_load_balance", None)
        if _cp_options_enable_load_balance is not None:
            _cp_options.enable_load_balance = False
        _cp_options_rotate_method = getattr(_cp_options, "rotate_method", None)
        if _cp_options_rotate_method is not None:
            _cp_options_rotate_method = _cp_options_rotate_method.value
            _cp_options.rotate_method = torch_ring_attention._RotateMethod.ALL_TO_ALL
    return _convert_to_f32, _cp_options_convert_to_f32, _cp_options_enable_load_balance, _cp_options_rotate_method


@torch.compiler.assume_constant_result
def _cleanup_ring_attn_func_forward(
    _convert_to_f32, _cp_options_convert_to_f32, _cp_options_enable_load_balance, _cp_options_rotate_method
):
    if _convert_to_f32 is not None:
        torch_ring_attention._convert_to_f32 = _convert_to_f32
    if _cp_options_convert_to_f32 is not None:
        torch_ring_attention._cp_options.convert_to_f32 = _cp_options_convert_to_f32
    if _cp_options_enable_load_balance is not None:
        torch_ring_attention._cp_options.enable_load_balance = _cp_options_enable_load_balance
    if _cp_options_rotate_method is not None:
        torch_ring_attention._cp_options.rotate_method = torch_ring_attention._RotateMethod(_cp_options_rotate_method)


class RingAttnFunc(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        query,
        key,
        value,
        attn_mask,
        dropout_p,
        is_causal,
        scale,
        mesh,
    ):
        assert _templated_ring_attention is not None, "RingAttnFunc requires a newer version of PyTorch"

        seq_dim_args = []
        if torch_version_check("__ge__", "2.6.0"):
            seq_dim_args = [query.ndim - 2]
        settings = _setup_ring_attn_func_forward()
        try:
            out, lse = _templated_ring_attention(
                mesh,
                *seq_dim_args,
                para_attn_ops.attention_forward_with_lse,
                query,
                key,
                value,
                attn_mask=attn_mask,
                dropout_p=dropout_p,
                is_causal=is_causal,
                scale=scale,
            )
        finally:
            _cleanup_ring_attn_func_forward(*settings)
        out = out.to(query.dtype)
        return out

    @staticmethod
    def backward(ctx, dout, *args):
        raise NotImplementedError("Backward pass for RingAttnFunc is not implemented")


def ring_attn_func(
    query,
    key,
    value,
    attn_mask=None,
    dropout_p=0.0,
    is_causal=False,
    *,
    scale=None,
    mesh=None,
):
    pg = DP.get_group(mesh)
    world_size = DP.get_world_size(pg)
    if world_size <= 1:
        return F.scaled_dot_product_attention(
            query,
            key,
            value,
            attn_mask=attn_mask,
            dropout_p=dropout_p,
            is_causal=is_causal,
            scale=scale,
        )

    assert attn_mask is None, "attn_mask is not supported in ring_attn_func when world_size > 1"

    return RingAttnFunc.apply(
        query,
        key,
        value,
        attn_mask,
        dropout_p,
        is_causal,
        scale,
        pg,
    )


def in_batch_attn_func(
    query,
    key,
    value,
    attn_mask=None,
    dropout_p=0.0,
    is_causal=False,
    *,
    scale=None,
):
    assert query.ndim == 4, "query must have 4 dimensions, got {}".format(query.ndim)
    assert key.ndim == 4, "key must have 4 dimensions, got {}".format(key.ndim)
    assert value.ndim == 4, "value must have 4 dimensions, got {}".format(value.ndim)
    assert attn_mask is None, "attn_mask is not supported in in_batch_attn_func"
    assert not is_causal, "is_causal is not supported in in_batch_attn_func"

    b, h, s_q, d_qk = query.shape
    _, _, s_kv, d_v = value.shape

    query = query.permute(1, 0, 2, 3).reshape(1, h, b * s_q, -1)
    key = key.permute(1, 0, 2, 3).reshape(1, h, b * s_kv, -1)
    value = value.permute(1, 0, 2, 3).reshape(1, h, b * s_kv, -1)

    out = F.scaled_dot_product_attention(
        query,
        key,
        value,
        attn_mask=attn_mask,
        dropout_p=dropout_p,
        is_causal=is_causal,
        scale=scale,
    )

    out = out.reshape(h, b, s_q, -1).permute(1, 0, 2, 3)
    return out


def _get_arg(args, kwargs, *field):
    if len(field) == 1:
        if isinstance(field, int):
            if field < len(args):
                return args[field]
            else:
                return None
        else:
            return kwargs.get(field[0])
    else:
        index, name = field
        if index < len(args):
            return args[index]
        else:
            return kwargs.get(name)


def _get_args(args, kwargs, *names):
    results = []
    for i, name in enumerate(names):
        results.append(_get_arg(args, kwargs, i, name))
    return results


class RingAttnMode(BaseTorchFunctionMode):
    disabled = False

    @torch.compiler.disable
    def __init__(self, mesh=None, *, skip_small_kv=False):
        super().__init__()
        self._mesh = mesh
        self._skip_small_kv = skip_small_kv

    def __torch_function__(self, func, types, args=(), kwargs=None):
        kwargs = {} if kwargs is None else kwargs

        if RingAttnMode.disabled:
            return base_handle_torch_function(func, types, args, kwargs)

        if func is F.scaled_dot_product_attention:
            if self._skip_small_kv:
                query, key = _get_args(args, kwargs, "query", "key")
                if query.shape[-2] > key.shape[-2]:
                    return base_handle_torch_function(func, types, args, kwargs)
            return self._call_ring_attn_func(*args, **kwargs)

        return base_handle_torch_function(func, types, args, kwargs)

    def _call_ring_attn_func(self, *args, **kwargs):
        mesh = self._mesh
        return ring_attn_func(*args, **kwargs, mesh=mesh)

    @classmethod
    @contextlib.contextmanager
    def disable(cls):
        old_disabled = cls._set_disabled(True)
        try:
            yield
        finally:
            cls._set_disabled(old_disabled)

    @classmethod
    @torch.compiler.disable
    def _set_disabled(cls, value):
        old_disabled = cls.disabled
        cls.disabled = value
        return old_disabled


class UlyssesAttnMode(BaseTorchFunctionMode):
    disabled = False

    @torch.compiler.disable
    def __init__(self, mesh=None, *, attn_func=None, skip_small_kv=False):
        super().__init__()
        self._mesh = mesh
        self._attn_func = attn_func
        self._skip_small_kv = skip_small_kv

    def __torch_function__(self, func, types, args=(), kwargs=None):
        kwargs = {} if kwargs is None else kwargs

        if UlyssesAttnMode.disabled:
            return base_handle_torch_function(func, types, args, kwargs)

        if func is F.scaled_dot_product_attention:
            if self._skip_small_kv:
                query, key = _get_args(args, kwargs, "query", "key")
                if query.shape[-2] > key.shape[-2]:
                    return base_handle_torch_function(func, types, args, kwargs)
            return self._call_ulysses_attn_func(*args, **kwargs)

        return base_handle_torch_function(func, types, args, kwargs)

    def _call_ulysses_attn_func(self, *args, **kwargs):
        mesh = self._mesh
        attn_func = self._attn_func
        return ulysses_attn_func(*args, **kwargs, mesh=mesh, attn_func=attn_func)

    @classmethod
    @contextlib.contextmanager
    def disable(cls):
        old_disabled = cls._set_disabled(True)
        try:
            yield
        finally:
            cls._set_disabled(old_disabled)

    @classmethod
    @torch.compiler.disable
    def _set_disabled(cls, value):
        old_disabled = cls.disabled
        cls.disabled = value
        return old_disabled


class UnifiedAttnMode(BaseTorchFunctionMode):
    disabled = False

    @torch.compiler.disable
    def __init__(self, mesh=None, *, skip_small_kv=False, attn_func=None):
        super().__init__()

        self._parallel_method = "ulysses"

        if mesh is None:
            self._ulysses_mesh = DP.get_default_group()
            self._ring_mesh = None
        else:
            if isinstance(mesh, dist.ProcessGroup):
                self._ulysses_mesh = mesh
                self._ring_mesh = None
            else:
                assert isinstance(mesh, dist.DeviceMesh), "mesh must be a ProcessGroup or DeviceMesh"

                if "ulysses" in mesh.mesh_dim_names:
                    self._ulysses_mesh = mesh["ulysses"]
                else:
                    self._ulysses_mesh = None
                if "ring" in mesh.mesh_dim_names:
                    self._ring_mesh = mesh["ring"]
                else:
                    self._ring_mesh = None

                assert (
                    self._ulysses_mesh is not None or self._ring_mesh is not None
                ), "mesh must have ulysses or ring dim"

        self._skip_small_kv = skip_small_kv
        self._attn_func = attn_func

    def __torch_function__(self, func, types, args=(), kwargs=None):
        kwargs = {} if kwargs is None else kwargs

        if UnifiedAttnMode.disabled:
            return base_handle_torch_function(func, types, args, kwargs)

        if func is F.scaled_dot_product_attention:
            if self._skip_small_kv:
                query, key = _get_args(args, kwargs, "query", "key")
                if query.shape[-2] > key.shape[-2]:
                    return base_handle_torch_function(func, types, args, kwargs)
            return self._call_unified_attn_func(*args, **kwargs)

        return base_handle_torch_function(func, types, args, kwargs)

    def _call_unified_attn_func(self, *args, **kwargs):
        func = F.scaled_dot_product_attention
        parallel_method = self._parallel_method
        if parallel_method == "ulysses":
            self._parallel_method = "ring"
            try:
                with self:
                    if self._ulysses_mesh is None:
                        out = func(*args, **kwargs)
                    else:
                        out = ulysses_attn_func(*args, **kwargs, mesh=self._ulysses_mesh)
            finally:
                self._parallel_method = parallel_method
        elif parallel_method == "ring":
            self._parallel_method = "none"
            try:
                with self:
                    if self._ring_mesh is None:
                        out = func(*args, **kwargs)
                    else:
                        out = ring_attn_func(*args, **kwargs, mesh=self._ring_mesh)
            finally:
                self._parallel_method = parallel_method
        elif parallel_method == "none":
            if get_force_dispatch_to_custom_ops():
                out = para_attn_ops.attention_forward(*args, **kwargs)
            else:
                attn_func = self._attn_func
                if attn_func is None:
                    attn_func = func
                out = attn_func(*args, **kwargs)
        else:
            raise ValueError(f"Unknown parallel method: {parallel_method}")

        return out

    @classmethod
    @contextlib.contextmanager
    def disable(cls):
        old_disabled = cls._set_disabled(True)
        try:
            yield
        finally:
            cls._set_disabled(old_disabled)

    @classmethod
    @torch.compiler.disable
    def _set_disabled(cls, value):
        old_disabled = cls.disabled
        cls.disabled = value
        return old_disabled


class InBatchAttnMode(BaseTorchFunctionMode):
    disabled = False

    @torch.compiler.disable
    def __init__(self):
        super().__init__()

    def __torch_function__(self, func, types, args=(), kwargs=None):
        kwargs = {} if kwargs is None else kwargs

        if InBatchAttnMode.disabled:
            return base_handle_torch_function(func, types, args, kwargs)

        if func is F.scaled_dot_product_attention:
            return in_batch_attn_func(*args, **kwargs)

        return base_handle_torch_function(func, types, args, kwargs)

    @classmethod
    @contextlib.contextmanager
    def disable(cls):
        old_disabled = cls._set_disabled(True)
        try:
            yield
        finally:
            cls._set_disabled(old_disabled)

    @classmethod
    @torch.compiler.disable
    def _set_disabled(cls, value):
        old_disabled = cls.disabled
        cls.disabled = value
        return old_disabled



import math


class TaylorSeer:
    def __init__(self, n_derivatives=2, warmup_steps=1, skip_interval_steps=1, compute_step_map=None):
        self.n_derivatives = n_derivatives
        self.ORDER = n_derivatives + 1
        self.warmup_steps = warmup_steps
        self.skip_interval_steps = skip_interval_steps
        self.compute_step_map = compute_step_map
        self.reset_cache()

    def reset_cache(self):
        self.state = {
            "dY_prev": [None] * self.ORDER,
            "dY_current": [None] * self.ORDER,
        }
        self.current_step = -1
        self.last_non_approximated_step = -1

    def should_compute_full(self, step=None):
        step = self.current_step if step is None else step
        if self.compute_step_map is not None:
            return self.compute_step_map[step]
        if step < self.warmup_steps or (step - self.warmup_steps + 1) % self.skip_interval_steps == 0:
            return True
        return False

    def approximate_derivative(self, Y):
        dY_current = [None] * self.ORDER
        dY_current[0] = Y
        window = self.current_step - self.last_non_approximated_step
        for i in range(self.n_derivatives):
            if self.state["dY_prev"][i] is not None and self.current_step > 1:
                dY_current[i + 1] = (dY_current[i] - self.state["dY_prev"][i]) / window
            else:
                break
        return dY_current

    def approximate_value(self):
        elapsed = self.current_step - self.last_non_approximated_step
        output = 0
        for i, derivative in enumerate(self.state["dY_current"]):
            if derivative is not None:
                output += (1 / math.factorial(i)) * derivative * (elapsed**i)
            else:
                break
        return output

    def mark_step_begin(self):
        self.current_step += 1

    def update(self, Y):
        self.state["dY_prev"] = self.state["dY_current"]
        self.state["dY_current"] = self.approximate_derivative(Y)
        self.last_non_approximated_step = self.current_step

    def step(self, Y):
        self.mark_step_begin()
        if self.should_compute_full():
            self.update(Y)
            return Y
        else:
            return self.approximate_value()


# Test case to predict y = x^2
if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np

    taylor_seer = TaylorSeer(n_derivatives=2, warmup_steps=2, skip_interval_steps=2)

    x_values = np.arange(0, 10, 0.1)
    y_true = x_values**2

    y_pred = []
    errors = []
    for x in x_values:
        y = x**2
        y_approx = taylor_seer.step(y)
        y_pred.append(y_approx)
        errors.append(abs(y - y_approx))

    plt.figure(figsize=(10, 5))
    plt.subplot(1, 2, 1)
    plt.plot(x_values, y_true, label="True $y=x^2$")
    plt.plot(x_values, y_pred, "--", label="TaylorSeer Approximation")
    plt.legend()
    plt.xlabel("x")
    plt.ylabel("y")
    plt.title("TaylorSeer Approximation Test")
    plt.grid()

    plt.subplot(1, 2, 2)
    plt.plot(x_values, errors, color="red", label="Absolute Error")
    plt.legend()
    plt.xlabel("x")
    plt.ylabel("Error")
    plt.title("Approximation Error")
    plt.grid()

    plt.tight_layout()
    plt.show()



import base64

import contextlib
import io
import time

import pytest
import pytest_html

import torch

if not torch.cuda.is_available():
    pytest.skip("CUDA is not available", allow_module_level=True)

from para_attn.distributed.mp_runner import MPDistRunner


class DiffusionPipelineRunner(MPDistRunner):
    @property
    def world_size(self):
        return torch.cuda.device_count()

    def mesh(self, device, max_batch_dim_size=None, max_ring_dim_size=None):
        from para_attn.context_parallel import init_context_parallel_mesh

        mesh = init_context_parallel_mesh(
            device, max_batch_dim_size=max_batch_dim_size, max_ring_dim_size=max_ring_dim_size
        )
        return mesh

    def new_pipe(self, dtype, device, rank):
        raise NotImplementedError

    def call_pipe(self, pipe, **kwargs):
        raise NotImplementedError

    @property
    def enable_vae_parallel(self):
        return False

    def _test_benchmark_pipe(self, dtype, device, parallelize, compile, max_batch_dim_size, max_ring_dim_size):
        if not parallelize and self.rank != 0:
            return

        with torch.cuda.device(self.rank) if device == "cuda" else contextlib.nullcontext():
            pipe = self.new_pipe(dtype, device)

            if parallelize:
                from para_attn.context_parallel.diffusers_adapters import parallelize_pipe

                mesh = self.mesh(device, max_batch_dim_size=max_batch_dim_size, max_ring_dim_size=max_ring_dim_size)
                parallelize_pipe(pipe, mesh=mesh)

                if self.enable_vae_parallel:
                    from para_attn.parallel_vae.diffusers_adapters import parallelize_vae

                    parallelize_vae(pipe.vae, mesh=mesh._flatten())

            if compile:
                if parallelize:
                    torch._inductor.config.reorder_for_compute_comm_overlap = True
                # If cudagraphs is enabled and parallelize is True, the test will hang indefinitely
                # after the last iteration.

                # TODO: Why CogVideoX fails with max-autotune-no-cudagraphs by outputing NAN?
                pipe.transformer = torch.compile(pipe.transformer)

            output_image, warmup_time, inference_time = None, None, None

            for i in range(2):
                torch.manual_seed(0)

                call_kwargs = {}
                if i == 0:
                    call_kwargs["num_inference_steps"] = 1
                begin = time.time()
                output = self.call_pipe(pipe, **call_kwargs)
                end = time.time()
                if i == 0:
                    warmup_time = end - begin
                    msg = f"Warm-up time taken: {warmup_time:.3f} seconds"
                else:
                    inference_time = end - begin
                    msg = f"Inference time taken: {inference_time:.3f} seconds"
                if self.rank == 0:
                    print(msg)
                if i != 0:
                    if hasattr(output, "images"):
                        output_image = output.images[0]
                    elif hasattr(output, "frames"):
                        video = output.frames[0]
                        output_image = video[0]

            return output_image, warmup_time, inference_time

    def process_task(self, *args, **kwargs):
        return self._test_benchmark_pipe(*args, **kwargs)


class _TestDiffusionPipeline:
    class Runner(DiffusionPipelineRunner):
        pass

    def test_benchmark_pipe(self, extras, dtype, device, parallelize, compile, max_batch_dim_size, max_ring_dim_size):
        with self.Runner().start() as runner:
            try:
                output_image, warmup_time, inference_time = runner(
                    (dtype, device, parallelize, compile, max_batch_dim_size, max_ring_dim_size),
                )
            except Exception as e:
                lines = str(e).split("\n")
                for line in lines:
                    if "is not divisible by world_size" in line:
                        pytest.skip(line)
                raise

        extras.append(pytest_html.extras.html(f"<div><p>Warm-up time taken: {warmup_time:.3f} seconds</p></div>"))
        extras.append(pytest_html.extras.html(f"<div><p>Inference time taken: {inference_time:.3f} seconds</p></div>"))
        image_bytes = io.BytesIO()
        output_image.save(image_bytes, format="JPEG")
        base64_image = base64.b64encode(image_bytes.getvalue()).decode("utf-8")
        extras.append(pytest_html.extras.jpg(base64_image, "Output image"))


class TestFluxPipeline(_TestDiffusionPipeline):
    class Runner(DiffusionPipelineRunner):
        def new_pipe(self, dtype, device):
            from diffusers import FluxPipeline

            pipe = FluxPipeline.from_pretrained(
                "black-forest-labs/FLUX.1-dev",
                torch_dtype=dtype,
            ).to(f"{device}:{self.rank}")
            return pipe

        def call_pipe(self, pipe, **kwargs):
            if "num_inference_steps" not in kwargs:
                kwargs["num_inference_steps"] = 28
            return pipe(
                "A cat holding a sign that says hello world",
                output_type="pil" if self.rank == 0 else "pt",
                **kwargs,
            )

        @property
        def enable_vae_parallel(self):
            return True

    @pytest.mark.parametrize("dtype", [torch.bfloat16])
    @pytest.mark.parametrize("device", ["cuda"])
    @pytest.mark.parametrize(
        "parallelize,compile,max_batch_dim_size,max_ring_dim_size",
        [
            # [False, False, None, None],
            # [False, True, None, None],
            [True, False, None, 2],
            [True, True, None, 2],
            # [True, False, None, None],
            # [True, True, None, None],
        ],
    )
    def test_benchmark_pipe(self, extras, dtype, device, parallelize, compile, max_batch_dim_size, max_ring_dim_size):
        super().test_benchmark_pipe(extras, dtype, device, parallelize, compile, max_batch_dim_size, max_ring_dim_size)


class TestMochiPipeline(_TestDiffusionPipeline):
    class Runner(DiffusionPipelineRunner):
        def new_pipe(self, dtype, device):
            from diffusers import MochiPipeline

            pipe = MochiPipeline.from_pretrained(
                "genmo/mochi-1-preview",
                torch_dtype=dtype,
            ).to(f"{device}:{self.rank}")
            pipe.enable_vae_tiling()
            return pipe

        def call_pipe(self, pipe, **kwargs):
            num_inference_steps = kwargs.get("num_inference_steps")
            if num_inference_steps is not None:
                kwargs["num_inference_steps"] = max(2, num_inference_steps)
            return pipe(
                "Close-up of a chameleon's eye, with its scaly skin changing color. Ultra high resolution 4k.",
                num_frames=84,
                output_type="pil" if self.rank == 0 else "pt",
                **kwargs,
            )

    @pytest.mark.parametrize("dtype", [torch.bfloat16])
    @pytest.mark.parametrize("device", ["cuda"])
    @pytest.mark.parametrize(
        "parallelize,compile,max_batch_dim_size,max_ring_dim_size",
        [
            # [False, False, None, None],
            # [False, True, None, None],
            [True, False, 2, None],
            [True, True, 2, None],
        ],
    )
    def test_benchmark_pipe(self, extras, dtype, device, parallelize, compile, max_batch_dim_size, max_ring_dim_size):
        super().test_benchmark_pipe(extras, dtype, device, parallelize, compile, max_batch_dim_size, max_ring_dim_size)


class TestCogVideoXPipeline(_TestDiffusionPipeline):
    class Runner(DiffusionPipelineRunner):
        def new_pipe(self, dtype, device):
            from diffusers import CogVideoXPipeline

            pipe = CogVideoXPipeline.from_pretrained(
                "THUDM/CogVideoX-5b",
                torch_dtype=dtype,
            ).to(f"{device}:{self.rank}")
            pipe.vae.enable_slicing()
            pipe.vae.enable_tiling()
            return pipe

        def call_pipe(self, pipe, **kwargs):
            if "num_inference_steps" not in kwargs:
                kwargs["num_inference_steps"] = 50
            prompt = "A panda, dressed in a small, red jacket and a tiny hat, sits on a wooden stool in a serene bamboo forest. The panda's fluffy paws strum a miniature acoustic guitar, producing soft, melodic tunes. Nearby, a few other pandas gather, watching curiously and some clapping in rhythm. Sunlight filters through the tall bamboo, casting a gentle glow on the scene. The panda's face is expressive, showing concentration and joy as it plays. The background includes a small, flowing stream and vibrant green foliage, enhancing the peaceful and magical atmosphere of this unique musical performance."
            return pipe(
                prompt=prompt,
                num_videos_per_prompt=1,
                num_frames=49,
                guidance_scale=6,
                output_type="pil" if self.rank == 0 else "pt",
                **kwargs,
            )

    @pytest.mark.parametrize("dtype", [torch.bfloat16])
    @pytest.mark.parametrize("device", ["cuda"])
    @pytest.mark.parametrize(
        "parallelize,compile,max_batch_dim_size,max_ring_dim_size",
        [
            # [False, False, None, None],
            # [False, True, None, None],
            [True, False, None, 2],
            [True, True, None, 2],
        ],
    )
    def test_benchmark_pipe(self, extras, dtype, device, parallelize, compile, max_batch_dim_size, max_ring_dim_size):
        super().test_benchmark_pipe(extras, dtype, device, parallelize, compile, max_batch_dim_size, max_ring_dim_size)


class TestWanVideoPipeline(_TestDiffusionPipeline):
    class Runner(DiffusionPipelineRunner):
        def new_pipe(self, dtype, device):
            from diffusers import WanPipeline

            # from diffusers.schedulers.scheduling_unipc_multistep import UniPCMultistepScheduler

            model_id = "Wan-AI/Wan2.1-T2V-1.3B-Diffusers"
            # model_id = "Wan-AI/Wan2.1-T2V-14B-Diffusers"
            pipe = WanPipeline.from_pretrained(model_id, torch_dtype=dtype)
            # pipe.scheduler = UniPCMultistepScheduler.from_config(pipe.scheduler.config, flow_shift=3.0)
            pipe.to(f"{device}:{self.rank}")
            return pipe

        def call_pipe(self, pipe, **kwargs):
            if "num_inference_steps" not in kwargs:
                kwargs["num_inference_steps"] = 30
            return pipe(
                prompt="An astronaut dancing vigorously on the moon with earth flying past in the background, hyperrealistic",
                negative_prompt="",
                height=480,
                width=832,
                num_frames=81,
                output_type="pil" if self.rank == 0 else "pt",
                **kwargs,
            )

    @pytest.mark.parametrize("dtype", [torch.float16])
    @pytest.mark.parametrize("device", ["cuda"])
    @pytest.mark.parametrize(
        "parallelize,compile,max_batch_dim_size,max_ring_dim_size",
        [
            # [False, False, None, None],
            # [False, True, None, None],
            [True, False, 2, None],
            [True, True, 2, None],
        ],
    )
    def test_benchmark_pipe(self, extras, dtype, device, parallelize, compile, max_batch_dim_size, max_ring_dim_size):
        super().test_benchmark_pipe(extras, dtype, device, parallelize, compile, max_batch_dim_size, max_ring_dim_size)


class TestWanImageToVideoPipeline(_TestDiffusionPipeline):
    class Runner(DiffusionPipelineRunner):
        def new_pipe(self, dtype, device):
            from diffusers import AutoencoderKLWan, WanImageToVideoPipeline
            from transformers import CLIPVisionModel

            model_id = "Wan-AI/Wan2.1-I2V-14B-480P-Diffusers"
            # model_id = "Wan-AI/Wan2.1-I2V-14B-720P-Diffusers"
            vae = AutoencoderKLWan.from_pretrained(model_id, subfolder="vae", torch_dtype=torch.float32)
            image_encoder = CLIPVisionModel.from_pretrained(
                model_id, subfolder="image_encoder", torch_dtype=torch.float32
            )
            pipe = WanImageToVideoPipeline.from_pretrained(
                model_id, vae=vae, image_encoder=image_encoder, torch_dtype=dtype
            )
            pipe.to(f"{device}:{self.rank}")
            return pipe

        def call_pipe(self, pipe, **kwargs):
            from diffusers.utils import load_image

            if "num_inference_steps" not in kwargs:
                kwargs["num_inference_steps"] = 30
            return pipe(
                prompt="A man with short gray hair plays a red electric guitar.",
                image=load_image(
                    "https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/diffusers/guitar-man.png"
                ),
                num_frames=129,
                output_type="pil" if self.rank == 0 else "pt",
                **kwargs,
            )

    @pytest.mark.parametrize("dtype", [torch.float16])
    @pytest.mark.parametrize("device", ["cuda"])
    @pytest.mark.parametrize(
        "parallelize,compile,max_batch_dim_size,max_ring_dim_size",
        [
            # [False, False, None, None],
            # [False, True, None, None],
            [True, False, None, None],
            [True, True, None, None],
        ],
    )
    def test_benchmark_pipe(self, extras, dtype, device, parallelize, compile, max_batch_dim_size, max_ring_dim_size):
        super().test_benchmark_pipe(extras, dtype, device, parallelize, compile, max_batch_dim_size, max_ring_dim_size)


class TestHunyuanVideoPipeline(_TestDiffusionPipeline):
    class Runner(DiffusionPipelineRunner):
        def new_pipe(self, dtype, device):
            from diffusers import HunyuanVideoPipeline, HunyuanVideoTransformer3DModel

            # [rank1]: RuntimeError: Expected mha_graph->execute(handle, variant_pack, workspace_ptr.get()).is_good() to be true, but got false.  (Could this error message be improved?  If so, please report an enhancement request to PyTorch.)
            # torch.backends.cuda.enable_cudnn_sdp(False)

            model_id = "tencent/HunyuanVideo"
            transformer = HunyuanVideoTransformer3DModel.from_pretrained(
                model_id,
                subfolder="transformer",
                torch_dtype=torch.bfloat16,
                revision="refs/pr/18",
            )
            pipe = HunyuanVideoPipeline.from_pretrained(
                model_id,
                transformer=transformer,
                torch_dtype=dtype,
                revision="refs/pr/18",
            ).to(f"{device}:{self.rank}")

            pipe.vae.enable_tiling(
                # Make it runnable on GPUs with 48GB memory
                # tile_sample_min_height=128,
                # tile_sample_stride_height=96,
                # tile_sample_min_width=128,
                # tile_sample_stride_width=96,
                # tile_sample_min_num_frames=32,
                # tile_sample_stride_num_frames=24,
            )

            # Fix OOM because of awful inductor lowering of attn_bias of _scaled_dot_product_efficient_attention
            # import para_attn
            #
            # para_attn.config.attention.force_dispatch_to_custom_ops = True

            return pipe

        def call_pipe(self, pipe, **kwargs):
            if "num_inference_steps" not in kwargs:
                kwargs["num_inference_steps"] = 30
            return pipe(
                prompt="A cat walks on the grass, realistic",
                height=720,
                width=1280,
                num_frames=129,
                output_type="pil" if self.rank == 0 else "pt",
                **kwargs,
            )

        @property
        def enable_vae_parallel(self):
            return True

    @pytest.mark.parametrize("dtype", [torch.float16])
    @pytest.mark.parametrize("device", ["cuda"])
    @pytest.mark.parametrize(
        "parallelize,compile,max_batch_dim_size,max_ring_dim_size",
        [
            # [False, False, None, None],
            # [False, True, None, None],
            [True, False, None, None],
            [True, True, None, None],
        ],
    )
    def test_benchmark_pipe(self, extras, dtype, device, parallelize, compile, max_batch_dim_size, max_ring_dim_size):
        super().test_benchmark_pipe(extras, dtype, device, parallelize, compile, max_batch_dim_size, max_ring_dim_size)


class TestHunyuanVideoImageToVideoPipeline(_TestDiffusionPipeline):
    class Runner(DiffusionPipelineRunner):
        def new_pipe(self, dtype, device):
            from diffusers import HunyuanVideoImageToVideoPipeline, HunyuanVideoTransformer3DModel

            # [rank1]: RuntimeError: Expected mha_graph->execute(handle, variant_pack, workspace_ptr.get()).is_good() to be true, but got false.  (Could this error message be improved?  If so, please report an enhancement request to PyTorch.)
            # torch.backends.cuda.enable_cudnn_sdp(False)

            model_id = "hunyuanvideo-community/HunyuanVideo-I2V"
            transformer = HunyuanVideoTransformer3DModel.from_pretrained(
                model_id,
                subfolder="transformer",
                torch_dtype=torch.bfloat16,
            )
            pipe = HunyuanVideoImageToVideoPipeline.from_pretrained(
                model_id,
                transformer=transformer,
                torch_dtype=dtype,
            ).to(f"{device}:{self.rank}")

            pipe.vae.enable_tiling(
                # Make it runnable on GPUs with 48GB memory
                # tile_sample_min_height=128,
                # tile_sample_stride_height=96,
                # tile_sample_min_width=128,
                # tile_sample_stride_width=96,
                # tile_sample_min_num_frames=32,
                # tile_sample_stride_num_frames=24,
            )

            # Fix OOM because of awful inductor lowering of attn_bias of _scaled_dot_product_efficient_attention
            # import para_attn
            #
            # para_attn.config.attention.force_dispatch_to_custom_ops = True

            return pipe

        def call_pipe(self, pipe, **kwargs):
            from diffusers.utils import load_image

            if "num_inference_steps" not in kwargs:
                kwargs["num_inference_steps"] = 30
            return pipe(
                prompt="A man with short gray hair plays a red electric guitar.",
                image=load_image(
                    "https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/diffusers/guitar-man.png"
                ),
                num_frames=129,
                output_type="pil" if self.rank == 0 else "pt",
                **kwargs,
            )

        @property
        def enable_vae_parallel(self):
            return True

    @pytest.mark.parametrize("dtype", [torch.float16])
    @pytest.mark.parametrize("device", ["cuda"])
    @pytest.mark.parametrize(
        "parallelize,compile,max_batch_dim_size,max_ring_dim_size",
        [
            # [False, False, None, None],
            # [False, True, None, None],
            [True, False, None, None],
            [True, True, None, None],
        ],
    )
    def test_benchmark_pipe(self, extras, dtype, device, parallelize, compile, max_batch_dim_size, max_ring_dim_size):
        super().test_benchmark_pipe(extras, dtype, device, parallelize, compile, max_batch_dim_size, max_ring_dim_size)


class FluxPipelineMPDistRunner(MPDistRunner):
    @property
    def world_size(self):
        return min(2, torch.cuda.device_count())

    def init_processor(self):
        from para_attn.context_parallel import init_context_parallel_mesh
        from para_attn.context_parallel.diffusers_adapters import parallelize_pipe
        from para_attn.parallel_vae.diffusers_adapters import parallelize_vae

        with torch.cuda.device(self.rank):
            from diffusers import FluxPipeline

            pipe = FluxPipeline.from_pretrained(
                "black-forest-labs/FLUX.1-dev",
                torch_dtype=torch.bfloat16,
            ).to(f"cuda:{self.rank}")

            mesh = init_context_parallel_mesh(
                pipe.device.type,
                max_ring_dim_size=2,
            )
            parallelize_pipe(
                pipe,
                mesh=mesh,
            )
            parallelize_vae(pipe.vae, mesh=mesh._flatten())

            self.pipe = pipe

    def process_task(self, *args, debug_raise_exception_processing=False, **kwargs):
        if debug_raise_exception_processing:
            raise RuntimeError("Debug exception")

        return self.pipe(
            "A cat holding a sign that says hello world",
            output_type="pil" if self.rank == 0 else "pt",
            **kwargs,
        )


class TestFluxPipelineMPDistRunner:
    def test_process_with_exception(self):
        def wrap_call(runner, *args, timeout=None, **kwargs):
            assert runner.is_alive() and runner.is_almost_idle()
            runner(args, kwargs, timeout=timeout)

        with FluxPipelineMPDistRunner().start() as runner:
            wrap_call(runner)
            with pytest.raises(Exception):
                wrap_call(runner, debug_raise_exception_processing=True)
            wrap_call(runner)



import functools

import torch
import torch.distributed as dist
from diffusers import AutoencoderKLHunyuanVideo
from diffusers.models.autoencoders.vae import DecoderOutput

import para_attn.primitives as DP
from para_attn.parallel_vae import init_parallel_vae_mesh


def send_tensor(tensor, dst, group):
    tensor = tensor.contiguous()
    dist.send_object_list([tensor.shape], dst=dst, group=group)
    dist.send(tensor, dst=dst, group=group)


def recv_tensor(src, group, device=None, dtype=None):
    objects = [None]
    dist.recv_object_list(objects, src=src, group=group)
    t = torch.empty(objects[0], device=device, dtype=dtype)
    dist.recv(t, src=src, group=group)
    return t


def parallelize_vae(vae: AutoencoderKLHunyuanVideo, *, mesh=None):
    mesh = init_parallel_vae_mesh(vae.device.type, mesh=mesh)

    group = DP.get_group(mesh)
    world_size = DP.get_world_size(group)
    rank = DP.get_rank(mesh)

    vae.enable_tiling()

    @functools.wraps(vae.__class__.tiled_encode)
    def new_tiled_encode(
        self,
        x: torch.Tensor,
        *args,
        **kwargs,
    ):
        batch_size, num_channels, num_frames, height, width = x.shape
        latent_height = height // self.spatial_compression_ratio
        latent_width = width // self.spatial_compression_ratio

        tile_latent_min_height = self.tile_sample_min_height // self.spatial_compression_ratio
        tile_latent_min_width = self.tile_sample_min_width // self.spatial_compression_ratio
        tile_latent_stride_height = self.tile_sample_stride_height // self.spatial_compression_ratio
        tile_latent_stride_width = self.tile_sample_stride_width // self.spatial_compression_ratio

        blend_height = tile_latent_min_height - tile_latent_stride_height
        blend_width = tile_latent_min_width - tile_latent_stride_width

        if hasattr(self, "tile_sample_min_height"):
            tile_sample_min_height = self.tile_sample_min_height
        else:
            tile_sample_min_height = self.tile_sample_min_size

        if hasattr(self, "tile_sample_min_width"):
            tile_sample_min_width = self.tile_sample_min_width
        else:
            tile_sample_min_width = self.tile_sample_min_size

        # Split x into overlapping tiles and encode them separately.
        # The tiles have an overlap to avoid seams between tiles.
        count = 0
        rows = []
        for i in range(0, height, self.tile_sample_stride_height):
            row = []
            for j in range(0, width, self.tile_sample_stride_width):
                if count % world_size == rank:
                    tile = x[:, :, :, i : i + tile_sample_min_height, j : j + tile_sample_min_width]
                    tile = self.encoder(tile)
                    tile = self.quant_conv(tile)
                else:
                    tile = None
                row.append(tile)
                count += 1
            rows.append(row)

        if rank == 0:
            count = 0
            for i in range(len(rows)):
                for j in range(len(rows[i])):
                    if count % world_size != rank:
                        rows[i][j] = recv_tensor(count % world_size, group, device=x.device, dtype=x.dtype)
                    count += 1
        else:
            for i in range(len(rows)):
                for j in range(len(rows[i])):
                    tile = rows[i][j]
                    if tile is not None:
                        send_tensor(tile, 0, group)

        if rank == 0:
            result_rows = []
            for i, row in enumerate(rows):
                result_row = []
                for j, tile in enumerate(row):
                    # blend the above tile and the left tile
                    # to the current tile and add the current tile to the result row
                    if i > 0:
                        tile = self.blend_v(rows[i - 1][j], tile, blend_height)
                    if j > 0:
                        tile = self.blend_h(row[j - 1], tile, blend_width)
                    result_row.append(tile[:, :, :, :tile_latent_stride_height, :tile_latent_stride_width])
                result_rows.append(torch.cat(result_row, dim=-1))

            enc = torch.cat(result_rows, dim=3)[:, :, :, :latent_height, :latent_width]
        else:
            enc = recv_tensor(rank - 1, group, device=x.device, dtype=x.dtype)
        if rank < world_size - 1:
            send_tensor(enc, rank + 1, group)
        return enc

    vae.tiled_encode = new_tiled_encode.__get__(vae)

    @functools.wraps(vae.__class__.tiled_decode)
    def new_tiled_decode(
        self,
        z: torch.Tensor,
        *args,
        return_dict: bool = False,
        **kwargs,
    ):
        batch_size, num_channels, num_frames, height, width = z.shape
        sample_height = height * self.spatial_compression_ratio
        sample_width = width * self.spatial_compression_ratio

        tile_latent_min_height = self.tile_sample_min_height // self.spatial_compression_ratio
        tile_latent_min_width = self.tile_sample_min_width // self.spatial_compression_ratio
        tile_latent_stride_height = self.tile_sample_stride_height // self.spatial_compression_ratio
        tile_latent_stride_width = self.tile_sample_stride_width // self.spatial_compression_ratio

        blend_height = self.tile_sample_min_height - self.tile_sample_stride_height
        blend_width = self.tile_sample_min_width - self.tile_sample_stride_width

        # Split z into overlapping tiles and decode them separately.
        # The tiles have an overlap to avoid seams between tiles.
        count = 0
        rows = []
        for i in range(0, height, tile_latent_stride_height):
            row = []
            for j in range(0, width, tile_latent_stride_width):
                if count % world_size == rank:
                    tile = z[:, :, :, i : i + tile_latent_min_height, j : j + tile_latent_min_width]
                    tile = self.post_quant_conv(tile)
                    decoded = self.decoder(tile)
                else:
                    decoded = None
                row.append(decoded)
                count += 1
            rows.append(row)

        if rank == 0:
            count = 0
            for i in range(len(rows)):
                for j in range(len(rows[i])):
                    if count % world_size != rank:
                        rows[i][j] = recv_tensor(count % world_size, group, device=z.device, dtype=z.dtype)
                    count += 1
        else:
            for i in range(len(rows)):
                for j in range(len(rows[i])):
                    decoded = rows[i][j]
                    if decoded is not None:
                        send_tensor(decoded, 0, group)

        if rank == 0:
            result_rows = []
            for i, row in enumerate(rows):
                result_row = []
                for j, tile in enumerate(row):
                    # blend the above tile and the left tile
                    # to the current tile and add the current tile to the result row
                    if i > 0:
                        tile = self.blend_v(rows[i - 1][j], tile, blend_height)
                    if j > 0:
                        tile = self.blend_h(row[j - 1], tile, blend_width)
                    result_row.append(tile[:, :, :, : self.tile_sample_stride_height, : self.tile_sample_stride_width])
                result_rows.append(torch.cat(result_row, dim=-1))

            dec = torch.cat(result_rows, dim=3)[:, :, :, :sample_height, :sample_width]
        else:
            dec = recv_tensor(rank - 1, group, device=z.device, dtype=z.dtype)
        if rank < world_size - 1:
            send_tensor(dec, rank + 1, group)

        if not return_dict:
            return (dec,)
        return DecoderOutput(dec, dec)

    vae.tiled_decode = new_tiled_decode.__get__(vae)

    return vae



import torch
import torch.distributed as dist
from diffusers import FluxPipeline

dist.init_process_group()

torch.cuda.set_device(dist.get_rank())

pipe = FluxPipeline.from_pretrained(
    "black-forest-labs/FLUX.1-dev",
    torch_dtype=torch.bfloat16,
).to("cuda")

from para_attn.context_parallel import init_context_parallel_mesh
from para_attn.context_parallel.diffusers_adapters import parallelize_pipe
from para_attn.parallel_vae.diffusers_adapters import parallelize_vae

mesh = init_context_parallel_mesh(
    pipe.device.type,
    max_ring_dim_size=2,
)
parallelize_pipe(
    pipe,
    mesh=mesh,
)
parallelize_vae(pipe.vae, mesh=mesh._flatten())

# from para_attn.first_block_cache.diffusers_adapters import apply_cache_on_pipe

# apply_cache_on_pipe(pipe)

# pipe.enable_model_cpu_offload(gpu_id=dist.get_rank())

# torch._inductor.config.reorder_for_compute_comm_overlap = True
# pipe.transformer = torch.compile(pipe.transformer, mode="max-autotune-no-cudagraphs")

image = pipe(
    "A cat holding a sign that says hello world",
    num_inference_steps=28,
    output_type="pil" if dist.get_rank() == 0 else "pt",
).images[0]

if dist.get_rank() == 0:
    print("Saving image to flux.png")
    image.save("flux.png")

dist.destroy_process_group()



import functools
from typing import Any, Dict, List, Optional, Union

import torch

from diffusers import DiffusionPipeline, WanTransformer3DModel
from diffusers.models.modeling_outputs import Transformer2DModelOutput
from diffusers.utils import logging, scale_lora_layers, unscale_lora_layers, USE_PEFT_BACKEND

import para_attn.primitives as DP
from para_attn.context_parallel import init_context_parallel_mesh
from para_attn.para_attn_interface import UnifiedAttnMode

logger = logging.get_logger(__name__)


def parallelize_transformer(transformer: WanTransformer3DModel, *, mesh=None):
    if getattr(transformer, "_is_parallelized", False):
        return transformer

    mesh = init_context_parallel_mesh(transformer.device.type, mesh=mesh)
    batch_mesh = mesh["batch"]
    seq_mesh = mesh["ring", "ulysses"]._flatten()

    @functools.wraps(transformer.__class__.forward)
    def new_forward(
        self,
        hidden_states: torch.Tensor,
        timestep: torch.LongTensor,
        encoder_hidden_states: torch.Tensor,
        encoder_hidden_states_image: Optional[torch.Tensor] = None,
        return_dict: bool = True,
        attention_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Union[torch.Tensor, Dict[str, torch.Tensor]]:
        if attention_kwargs is not None:
            attention_kwargs = attention_kwargs.copy()
            lora_scale = attention_kwargs.pop("scale", 1.0)
        else:
            lora_scale = 1.0

        if USE_PEFT_BACKEND:
            # weight the lora layers by setting `lora_scale` for each PEFT layer
            scale_lora_layers(self, lora_scale)
        else:
            if attention_kwargs is not None and attention_kwargs.get("scale", None) is not None:
                logger.warning("Passing `scale` via `attention_kwargs` when not using the PEFT backend is ineffective.")

        batch_size, num_channels, num_frames, height, width = hidden_states.shape
        p_t, p_h, p_w = self.config.patch_size
        post_patch_num_frames = num_frames // p_t
        post_patch_height = height // p_h
        post_patch_width = width // p_w

        rotary_emb = self.rope(hidden_states)

        hidden_states = self.patch_embedding(hidden_states)
        hidden_states = hidden_states.flatten(2).transpose(1, 2)

        temb, timestep_proj, encoder_hidden_states, encoder_hidden_states_image = self.condition_embedder(
            timestep, encoder_hidden_states, encoder_hidden_states_image
        )
        timestep_proj = timestep_proj.unflatten(1, (6, -1))

        if encoder_hidden_states_image is not None:
            encoder_hidden_states = torch.concat([encoder_hidden_states_image, encoder_hidden_states], dim=1)

        timestep_proj = DP.get_assigned_chunk(timestep_proj, dim=0, group=batch_mesh)
        temb = DP.get_assigned_chunk(temb, dim=0, group=batch_mesh)
        hidden_states = DP.get_assigned_chunk(hidden_states, dim=0, group=batch_mesh)
        hidden_states = DP.get_assigned_chunk(hidden_states, dim=-2, group=seq_mesh)
        encoder_hidden_states = DP.get_assigned_chunk(encoder_hidden_states, dim=0, group=batch_mesh)

        # rotary_emb is broadcast across the batch dimension
        rotary_emb = DP.get_assigned_chunk(rotary_emb, dim=-2, group=seq_mesh)

        with UnifiedAttnMode(mesh):
            hidden_states = self.call_transformer_blocks(
                hidden_states, encoder_hidden_states, timestep_proj, rotary_emb
            )

        # 5. Output norm, projection & unpatchify
        shift, scale = (self.scale_shift_table + temb.unsqueeze(1)).chunk(2, dim=1)
        hidden_states = (self.norm_out(hidden_states.float()) * (1 + scale) + shift).type_as(hidden_states)
        hidden_states = self.proj_out(hidden_states)

        hidden_states = DP.get_complete_tensor(hidden_states, dim=-2, group=seq_mesh)
        hidden_states = DP.get_complete_tensor(hidden_states, dim=0, group=batch_mesh)

        hidden_states = hidden_states.reshape(
            batch_size, post_patch_num_frames, post_patch_height, post_patch_width, p_t, p_h, p_w, -1
        )

        hidden_states = hidden_states.permute(0, 7, 1, 4, 2, 5, 3, 6)
        output = hidden_states.flatten(6, 7).flatten(4, 5).flatten(2, 3)

        if USE_PEFT_BACKEND:
            # remove `lora_scale` from each PEFT layer
            unscale_lora_layers(self, lora_scale)

        if not return_dict:
            return (output,)

        return Transformer2DModelOutput(sample=output)

    transformer.forward = new_forward.__get__(transformer)

    def call_transformer_blocks(self, hidden_states, encoder_hidden_states, *args, **kwargs):
        if torch.is_grad_enabled() and self.gradient_checkpointing:

            def create_custom_forward(module, return_dict=None):
                def custom_forward(*inputs):
                    if return_dict is not None:
                        return module(*inputs, return_dict=return_dict)
                    else:
                        return module(*inputs)

                return custom_forward

            ckpt_kwargs: Dict[str, Any] = {"use_reentrant": False}

            for block in self.blocks:
                hidden_states = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(block),
                    hidden_states,
                    encoder_hidden_states,
                    *args,
                    **kwargs,
                    **ckpt_kwargs,
                )

        else:
            for block in self.blocks:
                hidden_states = block(hidden_states, encoder_hidden_states, *args, **kwargs)

        return hidden_states

    transformer.call_transformer_blocks = call_transformer_blocks.__get__(transformer)

    transformer._is_parallelized = True

    return transformer


def parallelize_pipe(pipe: DiffusionPipeline, *, shallow_patch: bool = False, **kwargs):
    if not getattr(pipe, "_is_parallelized", False):
        original_call = pipe.__class__.__call__

        @functools.wraps(original_call)
        def new_call(self, *args, generator: Optional[Union[torch.Generator, List[torch.Generator]]] = None, **kwargs):
            if generator is None and getattr(self, "_is_parallelized", False):
                seed_t = torch.randint(0, torch.iinfo(torch.int64).max, [1], dtype=torch.int64, device=self.device)
                seed_t = DP.get_complete_tensor(seed_t, dim=0)
                seed_t = DP.get_assigned_chunk(seed_t, dim=0, idx=0)
                seed = seed_t.item()
                seed -= torch.iinfo(torch.int64).min
                generator = torch.Generator(self.device).manual_seed(seed)
            return original_call(self, *args, generator=generator, **kwargs)

        new_call._is_parallelized = True

        pipe.__class__.__call__ = new_call
        pipe.__class__._is_parallelized = True

    if not shallow_patch:
        parallelize_transformer(pipe.transformer, **kwargs)

    return pipe



import torch
import torch.distributed as dist
from diffusers import MochiPipeline
from diffusers.utils import export_to_video

dist.init_process_group()

torch.cuda.set_device(dist.get_rank())

pipe = MochiPipeline.from_pretrained(
    "genmo/mochi-1-preview",
    torch_dtype=torch.bfloat16,
).to("cuda")

from para_attn.context_parallel import init_context_parallel_mesh
from para_attn.context_parallel.diffusers_adapters import parallelize_pipe

parallelize_pipe(
    pipe,
    mesh=init_context_parallel_mesh(
        pipe.device.type,
        max_batch_dim_size=2,
    ),
)

# from para_attn.first_block_cache.diffusers_adapters import apply_cache_on_pipe

# apply_cache_on_pipe(pipe)

# Enable memory savings
# pipe.enable_model_cpu_offload(gpu_id=dist.get_rank())
pipe.enable_vae_tiling()

# torch._inductor.config.reorder_for_compute_comm_overlap = True
# pipe.transformer = torch.compile(pipe.transformer, mode="max-autotune-no-cudagraphs")

prompt = "Close-up of a chameleon's eye, with its scaly skin changing color. Ultra high resolution 4k."
video = pipe(
    prompt,
    num_frames=84,
    output_type="pil" if dist.get_rank() == 0 else "pt",
).frames[0]

if dist.get_rank() == 0:
    print("Saving video to mochi.mp4")
    export_to_video(video, "mochi.mp4", fps=30)

dist.destroy_process_group()



import functools
from typing import Any, Dict, List, Optional, Union

import torch
import torch.nn.functional as F
from diffusers import DiffusionPipeline, MochiTransformer3DModel
from diffusers.models.modeling_outputs import Transformer2DModelOutput
from diffusers.utils import logging, scale_lora_layers, unscale_lora_layers, USE_PEFT_BACKEND

import para_attn.primitives as DP
from para_attn.context_parallel import init_context_parallel_mesh
from para_attn.para_attn_interface import SparseKVAttnMode, UnifiedAttnMode

logger = logging.get_logger(__name__)  # pylint: disable=invalid-name


def parallelize_transformer(transformer: MochiTransformer3DModel, *, mesh=None):
    if getattr(transformer, "_is_parallelized", False):
        return transformer

    mesh = init_context_parallel_mesh(transformer.device.type, mesh=mesh)
    batch_mesh = mesh["batch"]
    seq_mesh = mesh["ring", "ulysses"]._flatten()

    @functools.wraps(transformer.__class__.forward)
    def new_forward(
        self,
        hidden_states: torch.Tensor,
        encoder_hidden_states: torch.Tensor,
        timestep: torch.LongTensor,
        encoder_attention_mask: torch.Tensor,
        attention_kwargs: Optional[Dict[str, Any]] = None,
        return_dict: bool = True,
    ) -> torch.Tensor:
        if attention_kwargs is not None:
            attention_kwargs = attention_kwargs.copy()
            lora_scale = attention_kwargs.pop("scale", 1.0)
        else:
            lora_scale = 1.0

        if USE_PEFT_BACKEND:
            # weight the lora layers by setting `lora_scale` for each PEFT layer
            scale_lora_layers(self, lora_scale)
        else:
            if attention_kwargs is not None and attention_kwargs.get("scale", None) is not None:
                logger.warning("Passing `scale` via `attention_kwargs` when not using the PEFT backend is ineffective.")

        batch_size, num_channels, num_frames, height, width = hidden_states.shape
        p = self.config.patch_size

        post_patch_height = height // p
        post_patch_width = width // p

        temb, encoder_hidden_states = self.time_embed(
            timestep,
            encoder_hidden_states,
            encoder_attention_mask,
            hidden_dtype=hidden_states.dtype,
        )

        hidden_states = hidden_states.permute(0, 2, 1, 3, 4).flatten(0, 1)
        hidden_states = self.patch_embed(hidden_states)
        hidden_states = hidden_states.unflatten(0, (batch_size, -1)).flatten(1, 2)

        image_rotary_emb = self.rope(
            self.pos_frequencies,
            num_frames,
            post_patch_height,
            post_patch_width,
            device=hidden_states.device,
            dtype=torch.float32,
        )

        latent_sequence_length = hidden_states.shape[1]
        latent_attention_mask = torch.ones(
            batch_size, 1, latent_sequence_length, device=hidden_states.device, dtype=torch.bool
        )  # [B, 1, N]
        attention_mask = torch.cat(
            [latent_attention_mask, encoder_attention_mask.unsqueeze(1).to(torch.bool)], dim=-1
        )  # [B, 1, N + M]

        world_size = DP.get_world_size(seq_mesh)

        temb = DP.get_assigned_chunk(temb, dim=0, group=batch_mesh)
        hidden_states = DP.get_assigned_chunk(hidden_states, dim=0, group=batch_mesh)
        hidden_states = DP.get_assigned_chunk(hidden_states, dim=-2, group=seq_mesh)
        encoder_hidden_states = DP.get_assigned_chunk(encoder_hidden_states, dim=0, group=batch_mesh)
        encoder_hidden_states = DP.get_assigned_chunk(encoder_hidden_states, dim=-2, group=seq_mesh)

        hidden_states_len = hidden_states.shape[-2]
        encoder_hidden_states_len = encoder_hidden_states.shape[-2]

        attention_mask = DP.get_assigned_chunk(attention_mask, dim=0, group=batch_mesh)
        attention_mask = attention_mask[..., :1, :]

        new_attention_mask = []
        for i in range(world_size):
            new_attention_mask.append(attention_mask[..., :, i * hidden_states_len : (i + 1) * hidden_states_len])
            new_attention_mask.append(
                attention_mask[
                    ...,
                    :,
                    world_size * hidden_states_len
                    + i * encoder_hidden_states_len : world_size * hidden_states_len
                    + (i + 1) * encoder_hidden_states_len,
                ]
            )
        new_attention_mask = torch.cat(new_attention_mask, dim=-1)
        attention_mask = new_attention_mask

        encoder_attention_mask = attention_mask
        del attention_mask

        freqs_cos, freqs_sin = image_rotary_emb

        def get_rotary_emb_chunk(freqs):
            freqs = DP.get_assigned_chunk(freqs, dim=0, group=seq_mesh)
            return freqs

        freqs_cos = get_rotary_emb_chunk(freqs_cos)
        freqs_sin = get_rotary_emb_chunk(freqs_sin)
        image_rotary_emb = (freqs_cos, freqs_sin)

        with SparseKVAttnMode(), UnifiedAttnMode(mesh):
            hidden_states, encoder_hidden_states = self.call_transformer_blocks(
                hidden_states, encoder_hidden_states, temb, encoder_attention_mask, image_rotary_emb
            )

        hidden_states = self.norm_out(hidden_states, temb)
        hidden_states = self.proj_out(hidden_states)

        hidden_states = DP.get_complete_tensor(hidden_states, dim=-2, group=seq_mesh)
        hidden_states = DP.get_complete_tensor(hidden_states, dim=0, group=batch_mesh)

        hidden_states = hidden_states.reshape(batch_size, num_frames, post_patch_height, post_patch_width, p, p, -1)
        hidden_states = hidden_states.permute(0, 6, 1, 2, 4, 3, 5)
        output = hidden_states.reshape(batch_size, -1, num_frames, height, width)

        if USE_PEFT_BACKEND:
            # remove `lora_scale` from each PEFT layer
            unscale_lora_layers(self, lora_scale)

        if not return_dict:
            return (output,)
        return Transformer2DModelOutput(sample=output)

    transformer.forward = new_forward.__get__(transformer)

    def call_transformer_blocks(self, hidden_states, encoder_hidden_states, *args, **kwargs):
        for i, block in enumerate(self.transformer_blocks):
            if torch.is_grad_enabled() and self.gradient_checkpointing:

                def create_custom_forward(module):
                    def custom_forward(*inputs):
                        return module(*inputs)

                    return custom_forward

                ckpt_kwargs: Dict[str, Any] = {"use_reentrant": False}
                hidden_states, encoder_hidden_states = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(block),
                    hidden_states,
                    encoder_hidden_states,
                    *args,
                    **kwargs,
                    **ckpt_kwargs,
                )
            else:
                hidden_states, encoder_hidden_states = block(
                    hidden_states,
                    encoder_hidden_states,
                    *args,
                    **kwargs,
                )

        return hidden_states, encoder_hidden_states

    transformer.call_transformer_blocks = call_transformer_blocks.__get__(transformer)

    def new_attn_processor_call(
        self,
        attn,
        hidden_states: torch.Tensor,
        encoder_hidden_states: torch.Tensor,
        attention_mask: torch.Tensor,
        image_rotary_emb: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        query = attn.to_q(hidden_states)
        key = attn.to_k(hidden_states)
        value = attn.to_v(hidden_states)

        query = query.unflatten(2, (attn.heads, -1))
        key = key.unflatten(2, (attn.heads, -1))
        value = value.unflatten(2, (attn.heads, -1))

        if attn.norm_q is not None:
            query = attn.norm_q(query)
        if attn.norm_k is not None:
            key = attn.norm_k(key)

        encoder_query = attn.add_q_proj(encoder_hidden_states)
        encoder_key = attn.add_k_proj(encoder_hidden_states)
        encoder_value = attn.add_v_proj(encoder_hidden_states)

        encoder_query = encoder_query.unflatten(2, (attn.heads, -1))
        encoder_key = encoder_key.unflatten(2, (attn.heads, -1))
        encoder_value = encoder_value.unflatten(2, (attn.heads, -1))

        if attn.norm_added_q is not None:
            encoder_query = attn.norm_added_q(encoder_query)
        if attn.norm_added_k is not None:
            encoder_key = attn.norm_added_k(encoder_key)

        if image_rotary_emb is not None:

            def apply_rotary_emb(x, freqs_cos, freqs_sin):
                x_even = x[..., 0::2].float()
                x_odd = x[..., 1::2].float()

                cos = (x_even * freqs_cos - x_odd * freqs_sin).to(x.dtype)
                sin = (x_even * freqs_sin + x_odd * freqs_cos).to(x.dtype)

                return torch.stack([cos, sin], dim=-1).flatten(-2)

            query = apply_rotary_emb(query, *image_rotary_emb)
            key = apply_rotary_emb(key, *image_rotary_emb)

        query, key, value = query.transpose(1, 2), key.transpose(1, 2), value.transpose(1, 2)
        encoder_query, encoder_key, encoder_value = (
            encoder_query.transpose(1, 2),
            encoder_key.transpose(1, 2),
            encoder_value.transpose(1, 2),
        )

        sequence_length = query.size(2)
        encoder_sequence_length = encoder_query.size(2)

        query = torch.cat([query, encoder_query], dim=2)
        key = torch.cat([key, encoder_key], dim=2)
        value = torch.cat([value, encoder_value], dim=2)

        hidden_states = F.scaled_dot_product_attention(
            query, key, value, attn_mask=attention_mask, dropout_p=0.0, is_causal=False
        )
        hidden_states = hidden_states.transpose(1, 2).flatten(2, 3)
        hidden_states = hidden_states.to(query.dtype)

        hidden_states, encoder_hidden_states = hidden_states.split_with_sizes(
            (sequence_length, encoder_sequence_length), dim=1
        )

        # linear proj
        hidden_states = attn.to_out[0](hidden_states)
        # dropout
        hidden_states = attn.to_out[1](hidden_states)

        if hasattr(attn, "to_add_out"):
            encoder_hidden_states = attn.to_add_out(encoder_hidden_states)

        return hidden_states, encoder_hidden_states

    transformer.transformer_blocks[0].attn1.processor.__class__.__call__ = new_attn_processor_call

    transformer._is_parallelized = True

    return transformer


def parallelize_pipe(pipe: DiffusionPipeline, *, shallow_patch: bool = False, **kwargs):
    if not getattr(pipe, "_is_parallelized", False):
        original_call = pipe.__class__.__call__

        @functools.wraps(original_call)
        def new_call(self, *args, generator: Optional[Union[torch.Generator, List[torch.Generator]]] = None, **kwargs):
            if generator is None and getattr(self, "_is_parallelized", False):
                seed_t = torch.randint(0, torch.iinfo(torch.int64).max, [1], dtype=torch.int64, device=self.device)
                seed_t = DP.get_complete_tensor(seed_t, dim=0)
                seed_t = DP.get_assigned_chunk(seed_t, dim=0, idx=0)
                seed = seed_t.item()
                seed -= torch.iinfo(torch.int64).min
                generator = torch.Generator(self.device).manual_seed(seed)
            return original_call(self, *args, generator=generator, **kwargs)

        new_call._is_parallelized = True

        pipe.__class__.__call__ = new_call
        pipe.__class__._is_parallelized = True

    if not shallow_patch:
        parallelize_transformer(pipe.transformer, **kwargs)

    return pipe



import functools

import torch
import torch.distributed as dist
from diffusers import AutoencoderKL
from diffusers.models.autoencoders.vae import DecoderOutput

import para_attn.primitives as DP
from para_attn.parallel_vae import init_parallel_vae_mesh


def send_tensor(tensor, dst, group):
    tensor = tensor.contiguous()
    dist.send_object_list([tensor.shape], dst=dst, group=group)
    dist.send(tensor, dst=dst, group=group)


def recv_tensor(src, group, device=None, dtype=None):
    objects = [None]
    dist.recv_object_list(objects, src=src, group=group)
    t = torch.empty(objects[0], device=device, dtype=dtype)
    dist.recv(t, src=src, group=group)
    return t


def parallelize_vae(vae: AutoencoderKL, *, mesh=None):
    mesh = init_parallel_vae_mesh(vae.device.type, mesh=mesh)

    group = DP.get_group(mesh)
    world_size = DP.get_world_size(group)
    rank = DP.get_rank(mesh)

    vae.enable_tiling()

    @functools.wraps(vae.__class__._tiled_encode)
    def new__tiled_encode(
        self,
        x: torch.Tensor,
        *args,
        **kwargs,
    ):
        overlap_size = int(self.tile_sample_min_size * (1 - self.tile_overlap_factor))
        blend_extent = int(self.tile_latent_min_size * self.tile_overlap_factor)
        row_limit = self.tile_latent_min_size - blend_extent

        # Split the image into 512x512 tiles and encode them separately.
        count = 0
        rows = []
        for i in range(0, x.shape[2], overlap_size):
            row = []
            for j in range(0, x.shape[3], overlap_size):
                if count % world_size == rank:
                    tile = x[:, :, i : i + self.tile_sample_min_size, j : j + self.tile_sample_min_size]
                    tile = self.encoder(tile)
                    if self.config.use_quant_conv:
                        tile = self.quant_conv(tile)
                else:
                    tile = None
                row.append(tile)
                count += 1
            rows.append(row)

        if rank == 0:
            count = 0
            for i in range(len(rows)):
                for j in range(len(rows[i])):
                    if count % world_size != rank:
                        rows[i][j] = recv_tensor(count % world_size, group, device=x.device, dtype=x.dtype)
                    count += 1
        else:
            for i in range(len(rows)):
                for j in range(len(rows[i])):
                    tile = rows[i][j]
                    if tile is not None:
                        send_tensor(tile, 0, group)

        if rank == 0:
            result_rows = []
            for i, row in enumerate(rows):
                result_row = []
                for j, tile in enumerate(row):
                    # blend the above tile and the left tile
                    # to the current tile and add the current tile to the result row
                    if i > 0:
                        tile = self.blend_v(rows[i - 1][j], tile, blend_extent)
                    if j > 0:
                        tile = self.blend_h(row[j - 1], tile, blend_extent)
                    result_row.append(tile[:, :, :row_limit, :row_limit])
                result_rows.append(torch.cat(result_row, dim=3))

            enc = torch.cat(result_rows, dim=2)
        else:
            enc = recv_tensor(rank - 1, group, device=x.device, dtype=x.dtype)
        if rank < world_size - 1:
            send_tensor(enc, rank + 1, group)
        return enc

    vae._tiled_encode = new__tiled_encode.__get__(vae)

    @functools.wraps(vae.__class__.tiled_decode)
    def new_tiled_decode(
        self,
        z: torch.Tensor,
        *args,
        return_dict: bool = False,
        **kwargs,
    ):
        overlap_size = int(self.tile_latent_min_size * (1 - self.tile_overlap_factor))
        blend_extent = int(self.tile_sample_min_size * self.tile_overlap_factor)
        row_limit = self.tile_sample_min_size - blend_extent

        # Split z into overlapping 64x64 tiles and decode them separately.
        # The tiles have an overlap to avoid seams between tiles.
        count = 0
        rows = []
        for i in range(0, z.shape[2], overlap_size):
            row = []
            for j in range(0, z.shape[3], overlap_size):
                if count % world_size == rank:
                    tile = z[:, :, i : i + self.tile_latent_min_size, j : j + self.tile_latent_min_size]
                    if self.config.use_post_quant_conv:
                        tile = self.post_quant_conv(tile)
                    decoded = self.decoder(tile)
                else:
                    decoded = None
                row.append(decoded)
                count += 1
            rows.append(row)

        if rank == 0:
            count = 0
            for i in range(len(rows)):
                for j in range(len(rows[i])):
                    if count % world_size != rank:
                        rows[i][j] = recv_tensor(count % world_size, group, device=z.device, dtype=z.dtype)
                    count += 1
        else:
            for i in range(len(rows)):
                for j in range(len(rows[i])):
                    decoded = rows[i][j]
                    if decoded is not None:
                        send_tensor(decoded, 0, group)

        if rank == 0:
            result_rows = []
            for i, row in enumerate(rows):
                result_row = []
                for j, tile in enumerate(row):
                    # blend the above tile and the left tile
                    # to the current tile and add the current tile to the result row
                    if i > 0:
                        tile = self.blend_v(rows[i - 1][j], tile, blend_extent)
                    if j > 0:
                        tile = self.blend_h(row[j - 1], tile, blend_extent)
                    result_row.append(tile[:, :, :row_limit, :row_limit])
                result_rows.append(torch.cat(result_row, dim=3))

            dec = torch.cat(result_rows, dim=2)
        else:
            dec = recv_tensor(rank - 1, group, device=z.device, dtype=z.dtype)
        if rank < world_size - 1:
            send_tensor(dec, rank + 1, group)
        if not return_dict:
            return (dec,)

        return DecoderOutput(sample=dec)

    vae.tiled_decode = new_tiled_decode.__get__(vae)

    return vae



import math

import torch.distributed as dist

import para_attn.primitives as DP


def init_context_parallel_mesh(
    device_type=None, *, mesh=None, max_batch_dim_size=None, max_ring_dim_size=None, max_ulysses_dim_size=None
):
    if mesh is not None:
        return mesh

    assert device_type is not None, "device must be provided if mesh is not provided"

    world_size = DP.get_world_size()
    if max_batch_dim_size is None:
        batch_dim_size = 1
    else:
        batch_dim_size = math.gcd(world_size, max_batch_dim_size)

    attn_world_size = world_size // batch_dim_size

    assert not (
        max_ring_dim_size is not None and max_ulysses_dim_size is not None
    ), "Only one of max_ulysses_dim_size and max_ring_dim_size can be set"

    if max_ulysses_dim_size is None:
        if max_ring_dim_size is None:
            ring_dim_size = 1
        else:
            ring_dim_size = math.gcd(attn_world_size, max_ring_dim_size)
        ulysses_dim_size = attn_world_size // ring_dim_size
    else:
        ulysses_dim_size = math.gcd(attn_world_size, max_ulysses_dim_size)
        ring_dim_size = attn_world_size // ulysses_dim_size

    mesh_shape = (batch_dim_size, ring_dim_size, ulysses_dim_size)
    return dist.init_device_mesh(device_type, mesh_shape, mesh_dim_names=("batch", "ring", "ulysses"))
